#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from abc import ABCMeta, abstractmethod
from typing import Callable, Optional, Sequence, Union

from pydicom.uid import UID

from wsidicom.codec import Encoder
from wsidicom.codec import Settings as EncoderSettings
from wsidicom.series import Labels, Overviews, Pyramids

"""A Target enables creating new instances."""


class Target(metaclass=ABCMeta):
    """A target should be initiated with a path or similar, and enable saving of
    instances into that path."""

    def __init__(
        self,
        uid_generator: Callable[..., UID],
        workers: int,
        chunk_size: int,
        include_pyramids: Optional[Sequence[int]] = None,
        include_levels: Optional[Sequence[int]] = None,
        add_missing_levels: bool = False,
        transcoding: Optional[Union[EncoderSettings, Encoder]] = None,
    ) -> None:
        """Initiate a target.

        Parameters
        ----------
        uid_generator: Callable[..., UID]
            Uid generator to use.
        workers: int
            Maximum number of thread workers to use.
        chunk_size: int
            Chunk size (number of tiles) to process at a time. Actual chunk
            size also depends on minimun_chunk_size from image_data.
        include_pyramids: Optional[Sequence[int]] = None
            Optional list indices (in present pyramids) to include.
        include_levels: Optional[Sequence[int]] = None
            Optional list indices (in present levels) to include, e.g. [0, 1]
            includes the two lowest levels. Negative indices can be used,
            e.g. [-1, -2] includes the two highest levels.
        add_missing_levels: bool = False
            If to add missing dyadic levels up to the single tile level.
        """
        self._uid_generator = uid_generator
        self._workers = workers
        self._chunk_size = chunk_size
        self._include_pyramids = include_pyramids
        self._include_levels = include_levels
        self._add_missing_levels = add_missing_levels
        self._instance_number = 0
        if isinstance(transcoding, EncoderSettings):
            self._transcoder = Encoder.create_for_settings(transcoding)
        else:
            self._transcoder = transcoding
        self.__enter__()

    @abstractmethod
    def save_pyramids(self, pyramids: Pyramids) -> None:
        """Should save the pyramids to the target."""
        raise NotImplementedError()

    @abstractmethod
    def save_labels(self, labels: Labels) -> None:
        """Should save the labels to the target."""
        raise NotImplementedError()

    @abstractmethod
    def save_overviews(self, overviews: Overviews) -> None:
        """Should save the overviews to the target."""
        raise NotImplementedError()

    @abstractmethod
    def close(self) -> None:
        """Should close any by the object opened resources."""
        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()

    @staticmethod
    def _is_included_level(
        level: int,
        present_levels: Sequence[int],
        allow_missing: bool,
        include_indices: Optional[Sequence[int]] = None,
    ) -> bool:
        """Return true if pyramid level is in included levels.

        Parameters
        ----------
        level: int
            Pyramid level to check.
        present_levels: Sequence[int]
            List of pyramid levels present.
        allow_missing: bool
            If to include missing levels (not in present_levels).
        include_indices: Optional[Sequence[int]] = None
            Optional list indices (in present levels) to include, e.g. [0, 1]
            includes the two lowest levels. Negative indices can be used,
            e.g. [-1, -2] includes the two highest levels. Default of None
            will not limit the selection. An empty sequence will excluded all
            levels.

        Returns
        -------
        bool
            True if level should be included.
        """
        if level not in present_levels:
            return allow_missing
        if include_indices is None:
            return True
        absolute_levels = [
            present_levels[level]
            for level in include_indices
            if -len(present_levels) <= level < len(present_levels)
        ]
        return level in absolute_levels



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Frame index for EOT, parsing the positions from the EOT."""

from pydicom.tag import Tag

from wsidicom.errors import WsiDicomFileError
from wsidicom.file.io.frame_index.offset_table import OffsetTable
from wsidicom.file.io.frame_index.offset_table_type import OffsetTableType
from wsidicom.tags import ExtendedOffsetTableLengthsTag, ExtendedOffsetTableTag


class Eot(OffsetTable):
    @property
    def offset_table_type(self) -> OffsetTableType:
        return OffsetTableType.EXTENDED

    @property
    def bytes_per_item(self) -> int:
        return 8

    @property
    def mode(self) -> str:
        return "<Q"

    def _get_index(self):
        table = self._read_table()
        return self._parse_table(table, self._pixels_start)

    def _get_pixels_start(self) -> int:
        pixel_data_or_eot_tag = Tag(self._file.read_tag())
        if pixel_data_or_eot_tag != ExtendedOffsetTableTag:
            raise WsiDicomFileError(str(self._file), "Expected EOT tag")
        self._file.read_tag_vr()
        eot_length = self._file.read_tag_length(True)
        self._file.seek(eot_length, 1)
        self._read_eot_lengths_tag()
        self._validate_pixel_data_start()
        bot_length = self._read_bot_length()
        if bot_length is not None:
            raise WsiDicomFileError(str(self._file), "Expected empty BOT")
        return self._file.tell()

    def _read_table(self) -> bytes:
        """Read extended table offset (EOT) and EOT lengths. Filepointer should be
        positionend to extended offset table.

        Returns
        -------
        bytes
            EOT in bytes.
        """
        eot_tag = Tag(self._file.read_tag())
        if eot_tag != ExtendedOffsetTableTag:
            raise ValueError(f"Expected ExtendedOffsetTable tag, got {eot_tag}")
        self._file.read_tag_vr()
        eot_length = self._read_eot_length()
        # Read the EOT into bytes
        eot = self._file.read(eot_length)
        # Read EOT lengths tag
        self._read_eot_lengths_tag()
        return eot

    def _read_eot_length(self) -> int:
        """Read the length of the extended table offset (EOT).

        Returns
        -------
        int
            EOT length.
        """
        EOT_BYTES = 8
        eot_length = self._file.read_tag_length(True)
        if eot_length == 0:
            raise WsiDicomFileError(
                str(self._file), "Expected Extended offset table present but empty"
            )
        elif eot_length % EOT_BYTES:
            raise WsiDicomFileError(
                str(self._file),
                "Extended offset table should be a multiple of " f"{EOT_BYTES} bytes",
            )
        return eot_length

    def _read_eot_lengths_tag(self):
        """Skip over the length of the extended table offset lengths tag."""
        eot_lenths_tag = self._file.read_tag()
        if eot_lenths_tag != ExtendedOffsetTableLengthsTag:
            raise WsiDicomFileError(
                str(self._file),
                "Expected Extended offset table lengths tag after reading "
                f"Extended offset table, found {eot_lenths_tag}",
            )
        self._file.read_tag_vr()
        length = self._file.read_tag_length(True)
        # Jump over EOT lengths for now
        self._file.seek(length, 1)



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from dataclasses import dataclass
from typing import Optional

from pydicom.uid import UID

from wsidicom.config import settings
from wsidicom.errors import WsiDicomStrictRequirementError

WSI_SOP_CLASS_UID = UID("1.2.840.10008.5.1.4.1.1.77.1.6")
ANN_SOP_CLASS_UID = UID("1.2.840.10008.5.1.4.1.1.91.1")
HTJPEG2000Lossless = UID("1.2.840.10008.1.2.4.201")
HTJPEG2000RPCLLossless = UID("1.2.840.10008.1.2.4.202")
HTJPEG2000 = UID("1.2.840.10008.1.2.4.203")


@dataclass(frozen=True)
class SlideUids:
    """Represents the UIDs that should be common for all files of a slide."""

    study_instance: UID
    series_instance: UID
    frame_of_reference: Optional[UID] = None

    def __post_init__(self) -> None:
        if settings.strict_uid_check and self.frame_of_reference is None:
            raise WsiDicomStrictRequirementError(
                "Frame of reference uid is missing and strict uid check is " "enabled"
            )

    def __str__(self) -> str:
        return (
            f"SlideUids study: {self.study_instance}, "
            f"series: {self.series_instance}, "
            f"frame of reference {self.frame_of_reference}"
        )

    def __eq__(self, other: "SlideUids") -> bool:
        if isinstance(other, SlideUids):
            return (
                self.study_instance == other.study_instance
                and self.series_instance == other.series_instance
                and (
                    self.frame_of_reference == other.frame_of_reference
                    or self.frame_of_reference is None
                    or other.frame_of_reference is None
                )
            )
        return NotImplemented

    def matches(self, other: "SlideUids") -> bool:
        if settings.strict_uid_check:
            return (
                self.study_instance == other.study_instance
                and self.frame_of_reference == other.frame_of_reference
            )

        return self.study_instance == other.study_instance


@dataclass(frozen=True)
class FileUids:
    """Represents the UIDs in a DICOM-file."""

    instance: UID
    concatenation: Optional[UID]
    slide: SlideUids

    @property
    def identifier(self) -> UID:
        """Return identifier for the instance the file belongs to. Either its
        own instance uid or, if not none, the concnatenation uid.

        Returns
        -------
        UID
            Identifier uid for the instance the file belongs to.
        """
        if self.concatenation is not None:
            return self.concatenation
        return self.instance

    def __eq__(self, other: "FileUids") -> bool:
        """Return true if concatenation uid is not none, matches other
        concatenation uid and base uids match.

        Parameters
        ----------
        other: FileUids
            File uids to match to.

        Returns
        -------
        bool
            True if concatenation is matching.
        """
        if isinstance(other, FileUids):
            return (
                self.concatenation is not None
                and self.concatenation == other.concatenation
                and self.slide == other.slide
            )
        return NotImplemented



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from pathlib import Path
from typing import Optional, Union

from PIL import Image as Pillow
from PIL.Image import Image
from pydicom.uid import UID, JPEGBaseline8Bit

from wsidicom.codec import Encoder
from wsidicom.geometry import (
    Point,
    Size,
    SizeMm,
)
from wsidicom.instance.image_data import ImageData
from wsidicom.metadata.image import ImageCoordinateSystem


class PillowImageData(ImageData):
    def __init__(self, image: Image):
        self._image = image.convert("RGB")
        encoder = Encoder.create(
            self.transfer_syntax,
            self.bits,
            self.photometric_interpretation,
        )
        super().__init__(encoder)

    @classmethod
    def from_file(cls, file: Union[str, Path]) -> "PillowImageData":
        image = Pillow.open(file)
        return cls(image)

    @property
    def transfer_syntax(self) -> UID:
        return JPEGBaseline8Bit

    @property
    def image_size(self) -> Size:
        return Size.from_tuple(self._image.size)

    @property
    def tile_size(self) -> Size:
        return self.image_size

    @property
    def pixel_spacing(self) -> Optional[SizeMm]:
        return None

    @property
    def samples_per_pixel(self) -> int:
        return 3

    @property
    def bits(self) -> int:
        return 8

    @property
    def lossy_compressed(self) -> bool:
        return True

    @property
    def photometric_interpretation(self) -> str:
        return "YBR_FULL_422"

    @property
    def image_coordinate_system(self) -> Optional[ImageCoordinateSystem]:
        return None

    def _get_decoded_tile(self, tile_point: Point, z: float, path: str) -> Image:
        if tile_point != Point(0, 0):
            raise ValueError("Can only get Point(0, 0) from non-tiled image.")
        return self._image

    def _get_encoded_tile(self, tile: Point, z: float, path: str) -> bytes:
        if tile != Point(0, 0):
            raise ValueError("Can only get Point(0, 0) from non-tiled image.")
        return self.encoder.encode(self._image)



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module with base marshmallow schemas that loads objects."""

import dataclasses
import logging
from abc import abstractmethod
from typing import Any, Dict, Generic, Type, TypeVar

from marshmallow import Schema, ValidationError, fields, post_load

LoadType = TypeVar("LoadType")


class LoadingSchema(Schema, Generic[LoadType]):
    """Schema for loading metadata"""

    @property
    @abstractmethod
    def load_type(self) -> Type[LoadType]:
        """Return the specimen type to use for deserialization."""
        raise NotImplementedError()

    @post_load
    def post_load(self, data: Dict[str, Any], **kwargs) -> LoadType:
        """Load object from dictionary."""
        return self.load_type(**data)


class DataclassLoadingSchema(LoadingSchema[LoadType]):
    """Schema for loading metadata"""

    @post_load
    def post_load(self, data: Dict[str, Any], **kwargs) -> LoadType:
        """Return a object of given load class using the defined dataclass fields."""
        assert dataclasses.is_dataclass(self.load_type)
        return self.load_type(
            **{
                field.name: data[field.name]
                for field in dataclasses.fields(self.load_type)
                if field.name in data
            }
        )


class DefaultOnValidationExceptionField(fields.Field):
    """Field that returns a default value if validation fails."""

    def __init__(self, inner: fields.Field, load_default, **kwargs):
        super().__init__(**kwargs)
        self.inner = inner
        self.load_default = load_default

    def _deserialize(self, value, attr, data, **kwargs):
        try:
            return self.inner._deserialize(value, attr, data, **kwargs)
        except ValidationError as exception:
            logging.warning(
                f"Could not deserialize {attr} using {self.inner} "
                f"due to exception {exception}"
                f"returning default value {self.load_default}"
            )
            return self.load_default

    def _serialize(self, value: Any, attr, obj, **kwargs):
        return self.inner._serialize(value, attr, obj, **kwargs)



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from pathlib import Path
from typing import List, Optional, Sequence, Tuple, Union

from PIL.Image import Image
from pydicom import Dataset
from pydicom.uid import UID

from wsidicom.errors import WsiDicomError, WsiDicomUidDuplicateError
from wsidicom.geometry import Size, SizeMm
from wsidicom.instance.dataset import ImageType, WsiDataset
from wsidicom.instance.image_data import ImageData
from wsidicom.instance.pillow_image_data import PillowImageData
from wsidicom.metadata.image import ImageCoordinateSystem
from wsidicom.uid import SlideUids


class WsiInstance:
    """Represents a level, label, or overview wsi image, containing image data
    and datasets with metadata."""

    def __init__(
        self, datasets: Union[WsiDataset, Sequence[WsiDataset]], image_data: ImageData
    ):
        """Create a WsiInstance from datasets with metadata and image data.

        Parameters
        ----------
        datasets: Union[WsiDataset, Sequence[WsiDataset]]
            Single dataset or list of datasets.
        image_data: ImageData
            Image data.
        """
        if not isinstance(datasets, Sequence):
            datasets = [datasets]
        self._datasets = datasets
        self._image_data = image_data
        self._identifier, self._uids = self._validate_instance(self.datasets)
        self._image_type = self.dataset.image_type

        if self.ext_depth_of_field:
            if self.ext_depth_of_field_planes is None:
                raise WsiDicomError("Instance Missing NumberOfFocalPlanes.")
            if self.ext_depth_of_field_plane_distance is None:
                raise WsiDicomError("Instance Missing DistanceBetweenFocalPlanes.")

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.dataset}, {self.image_data})"

    def __str__(self) -> str:
        return self.pretty_str()

    def pretty_str(self, indent: int = 0, depth: Optional[int] = None) -> str:
        string = f"default z: {self.default_z} " f"default path: { self.default_path}"
        if depth is not None:
            depth -= 1
            if depth < 0:
                return string
        string += " ImageData " + self.image_data.pretty_str(indent + 1, depth)
        return string

    @property
    def image_type(self) -> ImageType:
        """Return wsi type."""
        return self._image_type

    @property
    def datasets(self) -> Sequence[WsiDataset]:
        return self._datasets

    @property
    def dataset(self) -> WsiDataset:
        return self.datasets[0]

    @property
    def image_data(self) -> ImageData:
        return self._image_data

    @property
    def size(self) -> Size:
        """Return image size in pixels."""
        return self._image_data.image_size

    @property
    def tile_size(self) -> Size:
        """Return tile size in pixels."""
        return self._image_data.tile_size

    @property
    def mpp(self) -> Optional[SizeMm]:
        """Return pixel spacing in um/pixel."""
        if self.pixel_spacing is None:
            return None
        return self.pixel_spacing * 1000.0

    @property
    def pixel_spacing(self) -> Optional[SizeMm]:
        """Return pixel spacing in mm/pixel."""
        return self._image_data.pixel_spacing

    @property
    def mm_size(self) -> Optional[SizeMm]:
        """Return slide size in mm."""
        return self.dataset.mm_size

    @property
    def mm_depth(self) -> Optional[float]:
        """Return imaged depth in mm."""
        return self.dataset.mm_depth

    @property
    def slice_thickness(self) -> Optional[float]:
        """Return slice thickness."""
        return self.dataset.slice_thickness

    @property
    def slice_spacing(self) -> Optional[float]:
        """Return slice spacing."""
        return self.dataset.spacing_between_slices

    @property
    def focus_method(self) -> str:
        return self.dataset.focus_method

    @property
    def ext_depth_of_field(self) -> bool:
        return self.dataset.ext_depth_of_field

    @property
    def ext_depth_of_field_planes(self) -> Optional[int]:
        return self.dataset.ext_depth_of_field_planes

    @property
    def ext_depth_of_field_plane_distance(self) -> Optional[float]:
        return self.dataset.ext_depth_of_field_plane_distance

    @property
    def identifier(self) -> UID:
        """Return identifier (instance uid for single file instance or
        concatenation uid for multiple file instance)."""
        return self._identifier

    @property
    def default_z(self) -> float:
        return self._image_data.default_z

    @property
    def default_path(self) -> str:
        return self._image_data.default_path

    @property
    def focal_planes(self) -> List[float]:
        return self._image_data.focal_planes

    @property
    def optical_paths(self) -> List[str]:
        return self._image_data.optical_paths

    @property
    def tiled_size(self) -> Size:
        return self._image_data.tiled_size

    @property
    def uids(self) -> SlideUids:
        """Return base uids"""
        return self._uids

    @property
    def image_coordinate_system(self) -> Optional[ImageCoordinateSystem]:
        return self.image_data.image_coordinate_system

    @classmethod
    def create_label(
        cls, image: Union[Image, str, Path], base_dataset: Dataset
    ) -> "WsiInstance":
        """Create a label WsiInstance.

        Parameters
        ----------
        image: Union[Image, str, Path]
            Image or path to image.
        base_dataset: Dataset
            Base dataset to include.

        Returns
        -------
        WsiInstance
            Created label WsiInstance.
        """
        if isinstance(image, Image):
            image_data = PillowImageData(image)
        else:
            image_data = PillowImageData.from_file(image)
        return cls.create_instance(image_data, base_dataset, ImageType.LABEL)

    @classmethod
    def create_instance(
        cls, image_data: ImageData, base_dataset: Dataset, image_type: ImageType
    ) -> "WsiInstance":
        """Create WsiInstance from ImageData.

        Parameters
        ----------
        image_data: ImageData
            Image data and metadata.
        base_dataset: Dataset
            Base dataset to include.
        image_type: ImageType
            Type of instance to create.

        Returns
        -------
        WsiInstance
            Created WsiInstance.
        """
        instance_dataset = WsiDataset.create_instance_dataset(
            base_dataset, image_type, image_data
        )

        return cls(instance_dataset, image_data)

    @staticmethod
    def check_duplicate_instance(
        instances: Sequence["WsiInstance"], self: object
    ) -> None:
        """Check for duplicates in list of instances. Instances are duplicate
        if instance identifier (file instance uid or concatenation uid) match.
        Stops at first found duplicate and raises WsiDicomUidDuplicateError.

        Parameters
        ----------
        instances: Sequence['WsiInstance']
            List of instances to check.
        caller: Object
            Object that the instances belongs to.
        """
        instance_identifiers: List[str] = []
        for instance in instances:
            instance_identifier = instance.identifier
            if instance_identifier not in instance_identifiers:
                instance_identifiers.append(instance_identifier)
            else:
                raise WsiDicomUidDuplicateError(str(instance), str(self))

    def _validate_instance(
        self, datasets: Sequence[WsiDataset]
    ) -> Tuple[UID, SlideUids]:
        """Check that no files in instance are duplicate, that all files in
        instance matches (uid, type and size).
        Raises WsiDicomMatchError otherwise.
        Returns the matching file uid.

        Returns
        -------
        Tuple[UID, SlideUids]
            Instance identifier uid and base uids
        """
        WsiDataset.check_duplicate_dataset(datasets, self)

        base_dataset = datasets[0]
        for dataset in datasets[1:]:
            if not base_dataset.matches_instance(dataset):
                raise WsiDicomError("Datasets in instances does not match")
        return (
            base_dataset.uids.identifier,
            base_dataset.uids.slide,
        )

    def matches(self, other_instance: "WsiInstance") -> bool:
        """Return true if other instance is of the same group as self.

        Parameters
        ----------
        other_instance: WsiInstance
            Instance to check.

        Returns
        -------
        bool
            True if instances are of same group.

        """
        return (
            self.uids.matches(other_instance.uids)
            and self.size == other_instance.size
            and self.tile_size == other_instance.tile_size
            and self.image_type == other_instance.image_type
        )



#    Copyright 2021, 2022 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import math
from dataclasses import dataclass, replace
from typing import Callable, Iterable, Iterator, Sequence, Tuple, Union


@dataclass(frozen=True)
class SizeMm:
    width: float
    height: float

    def __mul__(self, factor: Union[int, float, "Size", "Point"]) -> "SizeMm":
        if isinstance(factor, (int, float)):
            return SizeMm(factor * self.width, factor * self.height)
        elif isinstance(factor, Size):
            return SizeMm(factor.width * self.width, factor.height * self.height)
        elif isinstance(factor, Point):
            return SizeMm(factor.x * self.width, factor.y * self.height)
        return NotImplemented

    def __truediv__(self, divider: Union[int, float, "SizeMm"]) -> "SizeMm":
        if isinstance(divider, (int, float)):
            return SizeMm(self.width / divider, self.height / divider)
        elif isinstance(divider, SizeMm):
            return SizeMm(self.width / divider.width, self.height / divider.height)
        return NotImplemented

    def __floordiv__(self, divider: Union[int, float, "SizeMm"]) -> "Size":
        if isinstance(divider, (int, float)):
            return Size(int(self.width / divider), int(self.height / divider))
        elif isinstance(divider, SizeMm):
            return Size(
                int(self.width / divider.width), int(self.height / divider.height)
            )
        return NotImplemented

    def to_int_tuple(self) -> Tuple[int, int]:
        return int(self.width), int(self.height)

    def to_tuple(self) -> Tuple[float, float]:
        return self.width, self.height

    @classmethod
    def from_tuple(cls, input: Union[Tuple[float, float], Sequence[float]]) -> "SizeMm":
        try:
            return cls(input[0], input[1])
        except IndexError:
            raise ValueError("input did not contain two values")


@dataclass(frozen=True)
class PointMm:
    x: float
    y: float

    def __mul__(self, factor: Union[int, float]) -> "PointMm":
        if isinstance(factor, (int, float)):
            return PointMm(factor * self.x, factor * self.y)
        return NotImplemented

    def __truediv__(self, divider: Union[int, float, "PointMm", "SizeMm"]) -> "PointMm":
        if isinstance(divider, (int, float)):
            return PointMm(self.x / divider, self.y / divider)
        elif isinstance(divider, PointMm):
            return PointMm(self.x / divider.x, self.y / divider.y)
        elif isinstance(divider, SizeMm):
            return PointMm(self.x / divider.width, self.y / divider.height)
        return NotImplemented

    def __floordiv__(self, divider: Union[int, float, "PointMm", "SizeMm"]) -> "Point":
        if isinstance(divider, (int, float)):
            return Point(int(self.x // divider), int(self.y // divider))
        elif isinstance(divider, PointMm):
            return Point(int(self.x // divider.x), int(self.y // divider.y))
        elif isinstance(divider, SizeMm):
            return Point(int(self.x // divider.width), int(self.y // divider.height))
        return NotImplemented

    def __add__(self, value: Union[int, float, "PointMm", "SizeMm"]) -> "PointMm":
        if isinstance(value, (int, float)):
            return PointMm(self.x + value, self.y + value)
        elif isinstance(value, SizeMm):
            return PointMm(self.x + value.width, self.y + value.height)
        elif isinstance(value, PointMm):
            return PointMm(self.x + value.x, self.y + value.y)
        return NotImplemented

    def __sub__(self, value: Union[int, float, "PointMm", "SizeMm"]) -> "PointMm":
        if isinstance(value, (int, float)):
            return PointMm(self.x - value, self.y - value)
        elif isinstance(value, SizeMm):
            return PointMm(self.x - value.width, self.y - value.height)
        elif isinstance(value, PointMm):
            return PointMm(self.x - value.x, self.y - value.y)
        return NotImplemented

    def __neg__(self) -> "PointMm":
        return PointMm(-self.x, -self.y)

    @classmethod
    def from_tuple(
        cls, input: Union[Tuple[float, float], Sequence[float]]
    ) -> "PointMm":
        try:
            return cls(input[0], input[1])
        except IndexError:
            raise ValueError("input did not contain two values")

    def to_tuple(self) -> Tuple[float, float]:
        return (self.x, self.y)


@dataclass(frozen=True)
class Size:
    width: int
    height: int

    def __neg__(self) -> "Size":
        return Size(-self.width, -self.height)

    def __sub__(self, value: Union[int, "Size", "Point"]) -> "Size":
        if isinstance(value, int):
            return Size(self.width - value, self.height - value)
        elif isinstance(value, Size):
            return Size(self.width - value.width, self.height - value.height)
        elif isinstance(value, Point):
            return Size(self.width - value.x, self.height - value.y)
        return NotImplemented

    def __add__(self, value: Union[int, "Size", "Point"]) -> "Size":
        if isinstance(value, int):
            return Size(self.width + value, self.height + value)
        elif isinstance(value, Size):
            return Size(self.width + value.width, self.height + value.height)
        elif isinstance(value, Point):
            return Size(self.width + value.x, self.height + value.y)
        return NotImplemented

    def __mul__(self, factor: Union[int, float, "Size", "Point"]) -> "Size":
        if isinstance(factor, (int, float)):
            return Size(int(factor * self.width), int(factor * self.height))
        elif isinstance(factor, Size):
            return Size(factor.width * self.width, factor.height * self.height)
        elif isinstance(factor, Point):
            return Size(factor.x * self.width, factor.y * self.height)
        return NotImplemented

    def __floordiv__(self, divider: Union[int, "Size", SizeMm]) -> "Size":
        if isinstance(divider, (int, float)):
            return Size(int(self.width / divider), int(self.height / divider))
        elif isinstance(divider, (Size, SizeMm)):
            return Size(
                int(self.width / divider.width), int(self.height / divider.height)
            )
        return NotImplemented

    def ceil_div(self, divider: Union[int, "Size", SizeMm]) -> "Size":
        if isinstance(divider, (int, float)):
            return Size(
                math.ceil(self.width / divider), math.ceil(self.height / divider)
            )
        elif isinstance(divider, (Size, SizeMm)):
            return Size(
                math.ceil(self.width / divider.width),
                math.ceil(self.height / divider.height),
            )
        return NotImplemented

    def __hash__(self) -> int:
        return hash((self.width, self.height))

    def all_less_than(self, item: "Size") -> bool:
        """If all dimensions is smaller than corresponding dimension in item."""
        return all(self._comparison_operation(item, int.__lt__))

    def all_less_than_or_equal(self, item: "Size") -> bool:
        """If all dimensions is smaller or equal to corresponding dimension in item."""
        return all(self._comparison_operation(item, int.__le__))

    def all_greater_than(self, item: "Size") -> bool:
        """If all dimensions is greater than to corresponding dimension in item."""
        return all(self._comparison_operation(item, int.__gt__))

    def all_greater_than_or_equal(self, item: "Size") -> bool:
        """If all dimensions is greater or equal to corresponding dimension in item."""
        return all(self._comparison_operation(item, int.__ge__))

    def any_less_than(self, item: "Size") -> bool:
        """If any dimension is smaller than corresponding dimension in item."""
        return any(self._comparison_operation(item, int.__lt__))

    def any_less_than_or_equal(self, item: "Size") -> bool:
        """If any dimension is smaller or equal to corresponding dimension in item."""
        return any(self._comparison_operation(item, int.__le__))

    def any_greater_than(self, item: "Size") -> bool:
        """If any dimension is greater than to corresponding dimension in item."""
        return any(self._comparison_operation(item, int.__gt__))

    def any_greater_than_or_equal(self, item: "Size") -> bool:
        """If any dimension is greater or equal to corresponding dimension in item."""
        return any(self._comparison_operation(item, int.__ge__))

    def _comparison_operation(
        self, item: "Size", operation: Callable[[int, int], bool]
    ) -> Iterable[bool]:
        return (
            operation(dimension, other_dimension)
            for (dimension, other_dimension) in zip(
                [self.width, self.height], [item.width, item.height]
            )
        )

    @classmethod
    def from_points(cls, point_1: "Point", point_2: "Point") -> "Size":
        return cls(point_2.x - point_1.x, point_2.y - point_1.y)

    def to_tuple(self) -> Tuple[int, int]:
        return (self.width, self.height)

    @classmethod
    def from_tuple(cls, input: Union[Tuple[int, int], Sequence[int]]) -> "Size":
        try:
            return cls(input[0], input[1])
        except IndexError:
            raise ValueError("input did not contain two values")

    @classmethod
    def max(cls, size_1: "Size", size_2: "Size") -> "Size":
        return cls(
            width=max(size_1.width, size_2.width),
            height=max(size_1.height, size_2.height),
        )

    def ceil(self) -> "Size":
        return Size(
            width=int(math.ceil(self.width)), height=int(math.ceil(self.height))
        )

    @property
    def area(self) -> int:
        return self.width * self.height


@dataclass(frozen=True)
class Point:
    x: int
    y: int

    def __str__(self) -> str:
        return f"{self.x},{self.y}"

    def __hash__(self) -> int:
        return hash((self.x, self.y))

    def __mul__(self, factor: Union[int, float, Size, "Point"]) -> "Point":
        if isinstance(factor, (int, float)):
            return Point(int(factor * self.x), int(factor * self.y))
        elif isinstance(factor, Size):
            return Point(factor.width * self.x, factor.height * self.y)
        elif isinstance(factor, Point):
            return Point(factor.x * self.x, factor.y * self.y)
        return NotImplemented

    def __floordiv__(
        self, divider: Union[int, float, "Point", Size, SizeMm]
    ) -> "Point":
        if isinstance(divider, (int, float)):
            return Point(int(self.x / divider), int(self.y / divider))
        elif isinstance(divider, Point):
            return Point(int(self.x / divider.x), int(self.y / divider.y))
        elif isinstance(divider, (Size, SizeMm)):
            return Point(int(self.x / divider.width), int(self.y / divider.height))
        return NotImplemented

    def ceil_div(self, divider: Union[int, float, Size, SizeMm]) -> "Point":
        if isinstance(divider, (int, float)):
            return Point(math.ceil(self.x / divider), math.ceil(self.y / divider))
        elif isinstance(divider, (Size, SizeMm)):
            return Point(
                math.ceil(self.x / divider.width), math.ceil(self.y / divider.height)
            )
        return NotImplemented

    def __mod__(self, divider: Union[Size, "Point"]) -> "Point":
        if isinstance(divider, Size):
            return Point(self.x % divider.width, self.y % divider.height)
        elif isinstance(divider, Point):
            return Point(self.x % divider.x, self.y % divider.y)
        return NotImplemented

    def __add__(self, value: Union[int, float, Size, "Point"]) -> "Point":
        if isinstance(value, (int, float)):
            return Point(int(self.x + value), int(self.y + value))
        elif isinstance(value, Size):
            return Point(self.x + value.width, self.y + value.height)
        elif isinstance(value, Point):
            return Point(self.x + value.x, self.y + value.y)
        return NotImplemented

    def __sub__(self, value: Union[int, float, Size, "Point"]) -> "Point":
        if isinstance(value, (int, float)):
            return Point(int(self.x - value), int(self.y - value))
        elif isinstance(value, Size):
            return Point(self.x - value.width, self.y - value.height)
        elif isinstance(value, Point):
            return Point(self.x - value.x, self.y - value.y)
        return NotImplemented

    @classmethod
    def max(cls, point_1: "Point", point_2: "Point") -> "Point":
        return cls(x=max(point_1.x, point_2.x), y=max(point_1.y, point_2.y))

    @classmethod
    def min(cls, point_1: "Point", point_2: "Point") -> "Point":
        return cls(x=min(point_1.x, point_2.x), y=min(point_1.y, point_2.y))

    def to_tuple(self) -> Tuple[int, int]:
        return (self.x, self.y)

    @classmethod
    def from_tuple(cls, input: Union[Tuple[int, int], Sequence[int]]) -> "Point":
        try:
            return cls(input[0], input[1])
        except IndexError:
            raise ValueError("input did not contain two values")


@dataclass(frozen=True)
class Region:
    position: Point
    size: Size

    @property
    def start(self) -> Point:
        return self.position

    @property
    def end(self) -> Point:
        end: Point = self.position + self.size
        return end

    def __mul__(self, value: int) -> "Region":
        if isinstance(value, int):
            return Region(position=self.position * value, size=self.size * value)
        return NotImplemented

    @property
    def box(self) -> Tuple[int, int, int, int]:
        return self.start.x, self.start.y, self.end.x, self.end.y

    @property
    def box_from_origin(self) -> Tuple[int, int, int, int]:
        return 0, 0, self.size.width, self.size.height

    def iterate_all(self) -> Iterator[Point]:
        return (
            Point(x, y)
            for y in range(self.start.y, self.end.y)
            for x in range(self.start.x, self.end.x)
        )

    def chunked_iterate_all(self, chunks: int) -> Iterator[Iterator[Point]]:
        points = list(self.iterate_all())
        points_count = len(points)
        chunk_size = math.ceil(points_count / chunks)
        for chunk_index in range(chunks):
            chunk_start = chunk_index * chunk_size
            chunk_end = min((chunk_index + 1) * chunk_size, points_count)
            yield (point for point in points[chunk_start:chunk_end])

    @classmethod
    def from_points(cls, point_1: "Point", point_2: "Point") -> "Region":
        return cls(
            position=point_1,
            size=Size(width=point_2.x - point_1.x, height=point_2.y - point_1.y),
        )

    @classmethod
    def from_tile(cls, tile: "Point", size: "Size") -> "Region":
        return cls(position=tile * size, size=size)

    def is_inside(self, test_region: "Region") -> bool:
        return bool(
            (self.start.x >= test_region.start.x)
            and (self.start.y >= test_region.start.y)
            and (self.end.x <= test_region.end.x)
            and (self.end.y <= test_region.end.y)
        )

    def crop(self, other_region: Union[Size, "Region"]) -> "Region":
        """Crop other_region to be inside this region. If the other_region is
        'Size', a region starting from Point(0, 0) is created.

        Parameters
        ----------
        other_region: Union[Size, 'Region']
            Region to be cropped.

        Returns
        -------
        Region
            Cropped region.
        """
        if isinstance(other_region, Size):
            other_region = Region(Point(0, 0), other_region)
        start = Point.min(Point.max(other_region.position, self.position), self.end)
        end = Point.max(Point.min(other_region.end, self.end), self.position)
        size = Size.from_points(start, end)
        return Region(position=start, size=size)

    def inside_crop(self, point: Point, tile_size: Size) -> "Region":
        """Crop point to be inside and with the same origin as this region.

        Parameters
        ----------
        point: Point
            Point to be cropped.

        Returns
        -------
        Region
            Cropped region.
        """
        tile_region = Region(position=point * tile_size, size=tile_size)
        cropped_tile_region = self.crop(tile_region)
        return replace(
            cropped_tile_region, position=cropped_tile_region.position % tile_size
        )

    def zoom(self, zoom: float) -> "Region":
        """Return center-zoomed region.

        Parameters
        ----------
        zoom: float
            The zoom level.

        Returns
        -------
        Region
            Zoomed region.
        """
        center = self.start + self.size // 2
        new_center = center * zoom
        new_start = new_center - self.size // 2
        return Region(new_start, self.size)


@dataclass(frozen=True)
class RegionMm:
    position: PointMm
    size: SizeMm

    # def __init__(self, position: PointMm, size: SizeMm):
    #     if size.width < 0:
    #         size.width = -size.width
    #         position.x -= size.width
    #     if size.height < 0:
    #         size.height = -size.height
    #         position.y -= size.height
    #     self.position = position
    #     self.size = size

    # @classmethod
    # def create(cls, position: PointMm, size: SizeMm) -> "RegionMm":
    #     if size.width < 0:
    #         size.width = -size.width
    #         position.x -= size.width
    #     if size.height < 0:
    #         size.height = -size.height
    #         position.y -= size.height
    #     return cls(position, size)

    @property
    def start(self) -> PointMm:
        return self.position

    @property
    def end(self) -> PointMm:
        end = self.position + self.size
        return end

    def __add__(self, value: PointMm) -> "RegionMm":
        if isinstance(value, PointMm):
            return RegionMm(self.position + value, self.size)
        return NotImplemented

    def __sub__(self, value: PointMm) -> "RegionMm":
        if isinstance(value, PointMm):
            return RegionMm(self.position - value, self.size)
        return NotImplemented

    def zoom(self, zoom: float) -> "RegionMm":
        """Return center-zoomed region.

        Parameters
        ----------
        zoom: float
            The zoom level.

        Returns
        -------
        RegionMm
            Zoomed region.
        """
        center = self.start + self.size / 2
        new_center = center * zoom
        new_start = new_center - self.size / 2
        return RegionMm(new_start, self.size)


class Orientation:
    def __init__(self, orientation: Tuple[float, float, float, float, float, float]):
        if orientation[0] != -orientation[4] or orientation[1] != orientation[3]:
            logging.warning(
                f"Orientation {orientation} is not "
                "orthogonal with equal lengths with column rotated 90 deg from row"
            )
        self._orientation = orientation
        self._transform = self._create_transform(orientation)
        self._reverse = self._create_reverse(self._transform)

    @property
    def rotation(self) -> float:
        return (
            (math.atan2(-self._orientation[0], self._orientation[3]) / math.pi)
            * 180
            % 360
        )

    @classmethod
    def from_rotation(cls, rotation: float) -> "Orientation":
        x = round(math.sin(rotation * math.pi / 180), 8)
        y = round(math.cos(rotation * math.pi / 180), 8)
        return cls((-x, y, 0, y, x, 0))

    @property
    def values(self) -> Tuple[float, float, float, float, float, float]:
        return self._orientation

    def apply_transform(self, point: PointMm):
        return PointMm(
            self._transform[0][0] * point.x + self._transform[0][1] * point.y,
            self._transform[1][0] * point.x + self._transform[1][1] * point.y,
        )

    def apply_reverse_transform(self, point: PointMm):
        return PointMm(
            self._reverse[0][0] * point.x + self._reverse[0][1] * point.y,
            self._reverse[1][0] * point.x + self._reverse[1][1] * point.y,
        )

    @staticmethod
    def _create_transform(
        orientation: Tuple[float, float, float, float, float, float]
    ) -> Tuple[Tuple[float, float], Tuple[float, float],]:
        return (
            (orientation[0], orientation[3]),
            (orientation[1], orientation[4]),
        )

    @staticmethod
    def _create_reverse(
        transform: Tuple[
            Tuple[float, float],
            Tuple[float, float],
        ]
    ) -> Tuple[Tuple[float, float], Tuple[float, float],]:
        determinant = 1 / (
            transform[0][0] * transform[1][1] - transform[0][1] * transform[1][0]
        )
        return (
            (determinant * transform[1][1], -determinant * transform[0][1]),
            (-determinant * transform[1][0], determinant * transform[0][0]),
        )



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""DICOM schema for Optical path model."""

from dataclasses import replace
import io
import struct
from enum import Enum
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
)

import numpy as np
from marshmallow import fields, post_load, pre_dump
from pydicom import Dataset
from pydicom.sr.coding import Code

from wsidicom.conceptcode import (
    IlluminationCode,
    IlluminationColorCode,
    ImagePathFilterCode,
    LenseCode,
    LightPathFilterCode,
)
from wsidicom.metadata.optical_path import (
    ConstantLutSegment,
    DiscreteLutSegment,
    ImagePathFilter,
    LightPathFilter,
    LinearLutSegment,
    Lut,
    LutDataType,
    LutSegment,
    Objectives,
    OpticalFilter,
    OpticalPath,
)
from wsidicom.metadata.schema.dicom.defaults import Defaults
from wsidicom.metadata.schema.dicom.fields import (
    CodeDicomField,
    DefaultingDicomField,
    FlattenOnDumpNestedDicomField,
    FloatDicomField,
    SingleCodeSequenceField,
    StringDicomField,
)
from wsidicom.metadata.schema.dicom.schema import (
    DicomSchema,
    LoadType,
    ModuleDicomSchema,
)


class LutSegmentType(Enum):
    DISCRETE = 0
    LINEAR = 1
    INDIRECT = 2


class LutDicomParser:
    @classmethod
    def from_dataset(cls, lut_sequence: Sequence[Dataset]) -> Optional[Lut]:
        """Read LUT from a DICOM lut sequence.

        Parameters
        ----------
        lut_sequence : Sequence[Dataset]
            DICOM optical path dataset.

        Returns
        -------
        Lut
            Parsed LUT.

        """
        dataset = lut_sequence[0]
        length, red_start, bits = dataset.RedPaletteColorLookupTableDescriptor
        _, green_start, _ = dataset.GreenPaletteColorLookupTableDescriptor
        _, blue_start, _ = dataset.BluePaletteColorLookupTableDescriptor
        if length == 0:
            length = 2**16
        bits = bits
        if bits == 8:
            data_type = np.uint8
        else:
            data_type = np.uint16

        segmented_keys = (
            "SegmentedRedPaletteColorLookupTableData",
            "SegmentedGreenPaletteColorLookupTableData",
            "SegmentedBluePaletteColorLookupTableData",
        )
        non_segmented_keys = (
            "RedPaletteColorLookupTableData",
            "GreenPaletteColorLookupTableData",
            "BluePaletteColorLookupTableData",
        )
        if all(key in dataset for key in segmented_keys):
            red = list(
                cls._parse_segments(
                    dataset.SegmentedRedPaletteColorLookupTableData, data_type
                )
            )
            green = list(
                cls._parse_segments(
                    dataset.SegmentedGreenPaletteColorLookupTableData, data_type
                )
            )
            blue = list(
                cls._parse_segments(
                    dataset.SegmentedBluePaletteColorLookupTableData, data_type
                )
            )

        elif all(key in dataset for key in non_segmented_keys):
            red: List[LutSegment] = [
                cls._parse_single_discrete_segment(
                    dataset.RedPaletteColorLookupTableData, data_type
                )
            ]
            green: List[LutSegment] = [
                cls._parse_single_discrete_segment(
                    dataset.GreenPaletteColorLookupTableData, data_type
                )
            ]
            blue: List[LutSegment] = [
                cls._parse_single_discrete_segment(
                    dataset.BluePaletteColorLookupTableData, data_type
                )
            ]

        else:
            raise ValueError(
                "Lookup table data or segmented lookup table data missing for one "
                "or more components."
            )
        for color, start in (
            (red, red_start),
            (green, green_start),
            (blue, blue_start),
        ):
            cls._add_start_and_end(start, length, color)

        return Lut(
            red=red,
            green=green,
            blue=blue,
            data_type=data_type,
        )

    @classmethod
    def _parse_segments(
        cls, segmented_lut_data: bytes, data_type: LutDataType
    ) -> Iterator[LutSegment]:
        """Parse segments from segmented lut data."""
        previous_segment_type: Optional[LutSegmentType] = None
        previous_segment_end_value: Optional[int] = None
        with io.BytesIO(segmented_lut_data) as buffer:
            data_type = cls._determine_correct_data_type(buffer, data_type)
            next_segment_type = cls._read_next_segment_type(buffer, data_type)
            while next_segment_type is not None:
                segment_type = next_segment_type
                if segment_type == LutSegmentType.DISCRETE:
                    length = cls._read_value(buffer, data_type)
                    values = cls._read_values(buffer, length, data_type)
                    next_segment_type = cls._read_next_segment_type(buffer, data_type)
                    if next_segment_type == LutSegmentType.LINEAR:
                        # If next segment is linear it will take over the last value.
                        previous_segment_end_value = values.pop()
                    if len(values) > 0:
                        yield DiscreteLutSegment(values)
                    previous_segment_type = segment_type

                elif segment_type == LutSegmentType.LINEAR:
                    length = cls._read_value(buffer, data_type)
                    end_value = cls._read_value(buffer, data_type)
                    if previous_segment_end_value is not None:
                        start_value = previous_segment_end_value
                        if previous_segment_type == LutSegmentType.DISCRETE:
                            # If the previous segment was discrete, this segment
                            # takes over its last value.
                            length += 1
                    else:
                        # The standard does allow the first segment to be a linear
                        # segment, but if it happens it is likely that the first value
                        # should be 0
                        start_value = 0
                    if start_value == end_value:
                        yield ConstantLutSegment(start_value, length)
                    else:
                        yield LinearLutSegment(start_value, end_value, length)
                    previous_segment_type = segment_type
                    next_segment_type = cls._read_next_segment_type(buffer, data_type)
                    previous_segment_end_value = end_value
                elif segment_type == LutSegmentType.INDIRECT:
                    raise NotImplementedError(
                        "Indirect segment types are not implemented."
                    )
                else:
                    raise ValueError("Unknown segment type.")

    @classmethod
    def _add_start_and_end(
        cls, start_length: int, total_length: int, segments: List[LutSegment]
    ):
        """Add start and end constant segments if needed."""
        start_segment = cls._create_start_segment(start_length, segments)
        end_segment = cls._create_end_segment(start_length, total_length, segments)
        if start_segment is not None:
            segments.insert(0, start_segment)
        if end_segment is not None:
            segments.append(end_segment)

    @staticmethod
    def _create_start_segment(
        start_length: int, segments: List[LutSegment]
    ) -> Optional[ConstantLutSegment]:
        """Create a start segment if needed."""
        if start_length == 0:
            return None
        first_segment = segments[0]
        if isinstance(first_segment, ConstantLutSegment):
            first_segment = replace(
                first_segment, length=first_segment.length + start_length
            )
            segments[0] = first_segment
            return None
        if isinstance(first_segment, DiscreteLutSegment):
            start_value = first_segment.values[0]
        elif isinstance(first_segment, LinearLutSegment):
            start_value = first_segment.start_value
        else:
            raise ValueError("Unknown segment type.")
        return ConstantLutSegment(start_value, start_length)

    @staticmethod
    def _create_end_segment(
        start_length, total_length: int, segments: List[LutSegment]
    ):
        """Create a end segment if needed."""
        length = start_length
        last_segment = segments[-1]
        length = sum(len(segment) for segment in segments)
        segment_length = total_length - length
        if segment_length < 0:
            raise ValueError("Got a negative length for last segment.")
        if segment_length == 0:
            return None
        if isinstance(last_segment, ConstantLutSegment):
            last_segment = replace(
                last_segment, length=last_segment.length + segment_length
            )
            segments[-1] = last_segment
            return None
        if isinstance(last_segment, DiscreteLutSegment):
            end_value = last_segment.values[-1]
        elif isinstance(last_segment, LinearLutSegment):
            end_value = last_segment.end_value
        else:
            raise ValueError("Unknown segment type.")
        return ConstantLutSegment(end_value, segment_length)

    @classmethod
    def _read_next_segment_type(
        cls, buffer: io.BytesIO, data_type: LutDataType
    ) -> Optional[LutSegmentType]:
        """
        Read next segment type from buffer.

        Return None if not enough data left to read.
        """
        try:
            return LutSegmentType(cls._read_value(buffer, data_type))
        except struct.error:
            return None

    @staticmethod
    def _read_value(buffer: io.BytesIO, data_type: LutDataType) -> int:
        """Read a single value from buffer."""
        if data_type == np.uint8:
            format = "<B"
        else:
            format = "<H"
        return struct.unpack(format, buffer.read(np.dtype(data_type).itemsize))[0]

    @staticmethod
    def _read_values(
        buffer: io.BytesIO, count: int, data_type: LutDataType
    ) -> List[int]:
        """Read multiple values from buffer."""
        if data_type == np.uint8:
            format = f'<{count*"B"}'
        else:
            format = f'<{count*"H"}'
        return list(
            struct.unpack(format, buffer.read(np.dtype(data_type).itemsize * count))
        )

    @classmethod
    def _parse_single_discrete_segment(
        cls, segment_data: bytes, data_type: LutDataType
    ) -> DiscreteLutSegment:
        """Read discrete segment from data."""
        length = len(segment_data) // np.dtype(data_type).itemsize
        if data_type == np.uint8:
            format = f'<{length*"B"}'
        else:
            format = f'<{length*"H"}'
        values = list(struct.unpack(format, segment_data))
        return DiscreteLutSegment(values)

    @classmethod
    def _determine_correct_data_type(
        cls, buffer: io.BytesIO, data_type: LutDataType
    ) -> LutDataType:
        """Determine correct data type for reading segment.

        The segment can either have 8- or 16-bits values. The first value indicates the
        segment type and should be 0, 1, or 2. The second value indicates a length that
        should be larger than 0.

        If reading 8 bits values as 16 bits the segment type will not be 0, 1, or 2, as
        the length value is also read. Then re-try reading as 8 bits.

        If reading 16 bit values as 8 bits the segment length will not be larger than 0,
        as the read value is the second half of the segment type value. Then re-try
        reading as 16 bits.
        """
        buffer.seek(0)
        try:
            cls._read_next_segment_type(buffer, data_type)
        except ValueError as exception:
            # Throws if first read data is not a valid segment type (0, 1, or 2.)
            if data_type == np.uint16:
                # Try reading the segment type as 8 bits
                return cls._determine_correct_data_type(buffer, np.uint8)
            raise ValueError("Failed to parse first segment type from data", exception)

        length = cls._read_value(buffer, data_type)
        if length > 0:
            # Length should be positive for all segment types
            buffer.seek(0)
            return data_type
        if data_type == np.uint8:
            # Try reading the segment as 16 bits.
            return cls._determine_correct_data_type(buffer, np.uint16)
        raise ValueError("Failed to parse first segment length from data.")


class LutDicomFormatter:
    @classmethod
    def to_dataset(cls, lut: Lut) -> List[Dataset]:
        """Convert lut into dataset."""
        dataset = Dataset()
        if lut.length == 2**16:
            length = 0
        else:
            length = lut.length
        red_start, red_data = cls._pack_segments(lut.red, lut.data_type)
        green_start, green_data = cls._pack_segments(lut.green, lut.data_type)
        blue_start, blue_data = cls._pack_segments(lut.blue, lut.data_type)
        dataset.RedPaletteColorLookupTableDescriptor = (length, red_start, lut.bits)
        dataset.GreenPaletteColorLookupTableDescriptor = (length, green_start, lut.bits)
        dataset.BluePaletteColorLookupTableDescriptor = (length, blue_start, lut.bits)
        dataset.SegmentedRedPaletteColorLookupTableData = red_data
        dataset.SegmentedGreenPaletteColorLookupTableData = green_data
        dataset.SegmentedBluePaletteColorLookupTableData = blue_data
        return [dataset]

    @classmethod
    def _pack_segments(
        cls,
        segments: Sequence[LutSegment],
        data_type: LutDataType,
    ) -> Tuple[int, bytes]:
        """Pack segments into bytes.

        Return start position (if constant first segment could be skipped) and bytes.
        """
        if data_type == np.uint8:
            data_format = "B"
        else:
            data_format = "H"
        previous_segment: Optional[LutSegment] = None
        end_index = len(segments) - 1
        start = 0
        with io.BytesIO() as buffer:
            for index, segment in enumerate(segments):
                if isinstance(segment, ConstantLutSegment):
                    if index == 0 and len(segments) > 1:
                        # Starting with constant segment and not only segment, we can
                        # skip it and set start of segmented data to length of segment.
                        start = segment.length
                    elif index != end_index or len(segments) == 1:
                        # Last constant segment can be skipped if not only segment.
                        cls._pack_constant_segment(
                            buffer,
                            segment,
                            previous_segment,
                            data_format,
                        )
                elif isinstance(segment, DiscreteLutSegment):
                    # Get next segment if not last segment.
                    if index < end_index:
                        next_segment = segments[index + 1]
                    else:
                        next_segment = None
                    cls._pack_discrete_segment(
                        buffer, segment, next_segment, data_format
                    )
                elif isinstance(segment, LinearLutSegment):
                    cls._pack_linear_segment(
                        buffer, segment, previous_segment, data_format
                    )
                else:
                    raise NotImplementedError()
                previous_segment = segment
            return start, buffer.getvalue()

    @classmethod
    def _pack_discrete_segment(
        cls,
        buffer: io.BytesIO,
        segment: DiscreteLutSegment,
        next_segment: Optional[LutSegment],
        data_format: str,
    ):
        values = list(segment.values)
        next_segment = next_segment
        if isinstance(next_segment, LinearLutSegment):
            values.append(next_segment.start_value)
        cls._pack_discrete_data(buffer, values, data_format)

    @classmethod
    def _pack_constant_segment(
        cls,
        buffer: io.BytesIO,
        segment: ConstantLutSegment,
        previous_segment: Optional[LutSegment],
        data_format: str,
    ):
        if not isinstance(previous_segment, DiscreteLutSegment):
            cls._pack_discrete_data(
                buffer,
                [segment.value],
                data_format,
            )
        cls._pack_linear_data(buffer, segment.length - 1, segment.value, data_format)

    @classmethod
    def _pack_linear_segment(
        cls,
        buffer: io.BytesIO,
        segment: LinearLutSegment,
        previous_segment: Optional[LutSegment],
        data_format: str,
    ):
        if not isinstance(previous_segment, DiscreteLutSegment):
            cls._pack_discrete_data(
                buffer,
                [segment.start_value],
                data_format,
            )
        cls._pack_linear_data(
            buffer, segment.length - 1, segment.end_value, data_format
        )

    @staticmethod
    def _pack_discrete_data(
        buffer: io.BytesIO,
        values: Sequence[int],
        data_format: str,
    ):
        """Pack discrete segment to buffer."""
        buffer.write(
            struct.pack(
                "<" + 2 * data_format,
                LutSegmentType.DISCRETE.value,
                len(values),
            )
        )
        for value in values:
            buffer.write(struct.pack("<" + data_format, value))

    @staticmethod
    def _pack_linear_data(
        buffer: io.BytesIO,
        length: int,
        end_value: int,
        data_format: str,
    ):
        """Pack linear segment to buffer."""
        buffer.write(
            struct.pack(
                "<" + 3 * data_format,
                LutSegmentType.LINEAR.value,
                length,
                end_value,
            )
        )


class LutDicomField(fields.Field):
    def _serialize(self, value: Optional[Lut], attr: Optional[str], obj: Any, **kwargs):
        if value is None:
            return None
        return LutDicomFormatter.to_dataset(value)

    def _deserialize(
        self,
        value: Optional[Sequence[Dataset]],
        attr: Optional[str],
        data: Optional[Dict[str, Any]],
        **kwargs,
    ) -> Optional[Lut]:
        if value is None or len(value) == 0:
            return None
        return LutDicomParser.from_dataset(value)


class FilterDicomSchema(DicomSchema[LoadType]):
    @pre_dump
    def pre_dump(self, filter: OpticalFilter, **kwargs):
        return {
            "filters": filter.filters,
            "nominal": filter.nominal,
            "filter_band": [filter.low_pass, filter.high_pass],
        }

    @post_load
    def post_load(self, data: Dict[str, Any], **kwargs):
        filter_band = data.pop("filter_band", None)
        if filter_band is not None:
            data["low_pass"] = filter_band[0]
            data["high_pass"] = filter_band[1]
        return super().post_load(data, **kwargs)


class LightPathFilterDicomSchema(FilterDicomSchema[LightPathFilter]):
    filters = fields.List(
        CodeDicomField(LightPathFilterCode),
        data_key="LightPathFilterTypeStackCodeSequence",
        allow_none=True,
    )
    nominal = fields.Integer(
        data_key="LightPathFilterPassThroughWavelength", allow_none=True
    )
    low_pass = fields.Integer(load_only=True, allow_none=True)
    high_pass = fields.Integer(load_only=True, allow_none=True)
    filter_band = fields.List(
        fields.Integer(),
        data_key="LightPathFilterPassBand",
    )

    @property
    def load_type(self) -> Type[LightPathFilter]:
        return LightPathFilter


class ImagePathFilterDicomSchema(FilterDicomSchema[ImagePathFilter]):
    filters = fields.List(
        CodeDicomField(ImagePathFilterCode),
        data_key="ImagePathFilterTypeStackCodeSequence",
        allow_none=True,
    )
    nominal = fields.Integer(
        data_key="ImagePathFilterPassThroughWavelength", allow_none=True
    )
    low_pass = fields.Integer(load_only=True, allow_none=True)
    high_pass = fields.Integer(load_only=True, allow_none=True)
    filter_band = fields.List(
        fields.Integer(),
        data_key="ImagePathFilterPassBand",
    )

    @property
    def load_type(self) -> Type[ImagePathFilter]:
        return ImagePathFilter


class ObjectivesSchema(DicomSchema[Objectives]):
    lenses = fields.List(
        CodeDicomField(LenseCode), data_key="LensesCodeSequence", allow_none=True
    )
    condenser_power = FloatDicomField(data_key="CondenserLensPower", allow_none=True)
    objective_power = FloatDicomField(data_key="ObjectiveLensPower", allow_none=True)
    objective_numerical_aperture = FloatDicomField(
        data_key="ObjectiveLensNumericalAperture", allow_none=True
    )

    @property
    def load_type(self) -> Type[Objectives]:
        return Objectives


class OpticalPathDicomSchema(ModuleDicomSchema[OpticalPath]):
    identifier = DefaultingDicomField(
        StringDicomField(),
        data_key="OpticalPathIdentifier",
        load_default=None,
        dump_default=Defaults.optical_path_identifier,
    )
    description = StringDicomField(data_key="OpticalPathDescription", load_default=None)
    illumination_types = DefaultingDicomField(
        fields.List(CodeDicomField(IlluminationCode)),
        data_key="IlluminationTypeCodeSequence",
        dump_default=[Defaults.illumination_type],
    )
    illumination_wavelength = fields.Integer(
        data_key="IlluminationWaveLength", load_default=None
    )
    illumination_color_code = SingleCodeSequenceField(
        IlluminationColorCode,
        data_key="IlluminationColorCodeSequence",
        load_default=None,
    )

    icc_profile = fields.Raw(data_key="ICCProfile", load_default=None)
    lut = LutDicomField(data_key="PaletteColorLookupTableSequence", load_default=None)
    light_path_filter = FlattenOnDumpNestedDicomField(
        LightPathFilterDicomSchema(), load_default=None
    )
    image_path_filter = FlattenOnDumpNestedDicomField(
        ImagePathFilterDicomSchema(), load_default=None
    )
    objective = FlattenOnDumpNestedDicomField(ObjectivesSchema(), load_default=None)

    @property
    def load_type(self) -> Type[OpticalPath]:
        return OpticalPath

    @pre_dump
    def pre_dump(self, optical_path: OpticalPath, **kwargs):
        fields = {
            "identifier": optical_path.identifier,
            "illumination_types": optical_path.illumination_types,
            "light_path_filter": optical_path.light_path_filter,
            "image_path_filter": optical_path.image_path_filter,
            "objective": optical_path.objective,
            "lut": optical_path.lut,
        }
        if optical_path.description is not None:
            fields["description"] = optical_path.description
        if optical_path.icc_profile is not None:
            fields["icc_profile"] = optical_path.icc_profile
        if isinstance(optical_path.illumination, float):
            fields["illumination_wavelength"] = optical_path.illumination
        if isinstance(optical_path.illumination, Code):
            fields["illumination_color_code"] = optical_path.illumination
        else:
            fields["illumination_color_code"] = Defaults.illumination
        return fields

    @post_load
    def post_load(self, data: Dict[str, Any], **kwargs):
        illumination_wavelength = data.pop("illumination_wavelength", None)
        illumination_color_code = data.pop("illumination_color_code", None)
        if illumination_wavelength is not None:
            data["illumination"] = illumination_wavelength
        elif illumination_color_code is not None:
            data["illumination"] = illumination_color_code
        return super().post_load(data, **kwargs)

    @property
    def module_name(self) -> str:
        return "optical_path"



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module with imports for optional codec packages."""


try:
    from rle.utils import decode_frame as rle_decode_frame
    from rle.utils import encode_frame as rle_encode_frame

    PYLIBJPEGRLE_AVAILABLE = True
except ImportError:
    PYLIBJPEGRLE_AVAILABLE = False
    rle_decode_frame = None
    rle_encode_frame = None

try:
    from imagecodecs import (
        JPEG2K,
        JPEG8,
        JPEGLS,
        dicomrle_decode,
        jpeg2k_decode,
        jpeg2k_encode,
        jpeg8_decode,
        jpeg8_encode,
        jpegls_decode,
        jpegls_encode,
        packbits_encode,
    )

    IMAGE_CODECS_AVAILABLE = True

except ImportError:
    IMAGE_CODECS_AVAILABLE = False
    JPEG2K = None
    JPEG8 = None
    JPEGLS = None
    jpeg2k_decode = None
    jpeg8_decode = None
    jpegls_decode = None
    jpeg2k_encode = None
    jpeg8_encode = None
    jpegls_encode = None
    packbits_encode = None
    dicomrle_decode = None

try:
    from jpeg_ls import decode as pylibjpeg_ls_decode
    from jpeg_ls import encode_array as pylibjpeg_ls_encode

    PYLIBJPEGLS_AVAILABLE = True

except ImportError:
    pylibjpeg_ls_decode = None
    pylibjpeg_ls_encode = None
    PYLIBJPEGLS_AVAILABLE = False


try:
    from openjpeg.utils import decode as pylibjpeg_openjpeg_decode
    from openjpeg.utils import encode as pylibjpeg_openjpeg_encode

    PYLIBJPEGOPENJPEG_AVAILABLE = True

except ImportError:
    pylibjpeg_openjpeg_decode = None
    pylibjpeg_openjpeg_encode = None
    PYLIBJPEGOPENJPEG_AVAILABLE = False



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from dataclasses import dataclass
from typing import Any, ClassVar, Dict, List, Optional, Type, TypeVar

from pydicom.dataset import Dataset
from pydicom.sequence import Sequence as DicomSequence
from pydicom.sr.codedict import codes
from pydicom.sr.coding import Code

ConceptCodeType = TypeVar("ConceptCodeType", bound="ConceptCode")
CidConceptCodeType = TypeVar("CidConceptCodeType", bound="CidConceptCode")


@dataclass(frozen=True)
class ConceptCode:
    """Help functions for handling SR codes.
    Provides functions for converting between Code and dicom dataset.
    """

    value: str
    scheme_designator: str
    meaning: str
    scheme_version: Optional[str] = None
    sequence_name: ClassVar[str]

    def __hash__(self):
        return hash(
            (self.value, self.scheme_designator, self.meaning, self.scheme_version)
        )

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Code):
            return self.code == other
        if isinstance(other, ConceptCode):
            return self.code == other.code
        return NotImplemented

    @property
    def code(self) -> Code:
        return Code(
            value=self.value,
            scheme_designator=self.scheme_designator,
            meaning=self.meaning,
            scheme_version=self.scheme_version,
        )

    @classmethod
    def from_code(cls: Type[ConceptCodeType], code: Code) -> ConceptCodeType:
        return cls(
            value=code.value,
            scheme_designator=code.scheme_designator,
            meaning=code.meaning,
            scheme_version=code.scheme_version,
        )

    def to_ds(self) -> Dataset:
        """Codes code into DICOM dataset.

        Returns
        -------
        Dataset
            Dataset of code.

        """
        ds = Dataset()
        ds.CodeValue = self.value
        ds.CodingSchemeDesignator = self.scheme_designator
        ds.CodeMeaning = self.meaning
        if self.scheme_version is not None:
            ds.CodeSchemeVersion = self.scheme_version
        return ds

    def insert_into_ds(self, ds: Dataset) -> Dataset:
        """Codes and insert object into sequence in dataset.

        Parameters
        ----------
        ds: Dataset
           Dataset to insert into.

        Returns
        -------
        Dataset
            Dataset with object inserted.

        """
        # Append if sequence already set otherwise create
        try:
            sequence = getattr(ds, self.sequence_name)
            sequence.append(self.to_ds())
        except AttributeError:
            setattr(ds, self.sequence_name, DicomSequence([self.to_ds()]))
        return ds

    @classmethod
    def _from_ds(
        cls: Type[ConceptCodeType],
        ds: Dataset,
    ) -> Optional[List[ConceptCodeType]]:
        """Return list of ConceptCode from sequence in dataset.

        Parameters
        ----------
        ds: Dataset
            Datasete containing code sequence.

        Returns
        -------
        Optional[List[ConceptCodeType]
            Codes created from sequence in dataset.

        """
        if cls.sequence_name not in ds:
            return None
        return [
            cls(
                value=code_ds.CodeValue,
                scheme_designator=code_ds.CodingSchemeDesignator,
                meaning=code_ds.CodeMeaning,
                scheme_version=getattr(code_ds, "CodeSchemeVersion", None),
            )
            for code_ds in getattr(ds, cls.sequence_name)
        ]


class SingleConceptCode(ConceptCode):
    """Code for concepts  # type: ignore that only allow a single item"""

    @classmethod
    def from_ds(cls: Type[ConceptCodeType], ds: Dataset) -> Optional["ConceptCodeType"]:
        """Return measurement code for value. Value can be a code meaning (str)
        or a DICOM dataset containing the code.

        Parameters
        ----------
        value: Dataset
            The dataset for creating the code.

        Returns
        -------
        Optional[ConceptCodeType]
            Concept code created from value.

        """
        codes = cls._from_ds(ds)
        if codes is None:
            return None
        return codes[0]


class MultipleConceptCode(ConceptCode):
    """Code for concepts  # type: ignore that allow multiple items"""

    @classmethod
    def from_ds(cls: Type[ConceptCodeType], ds: Dataset) -> List[ConceptCodeType]:
        """Return measurement code for value. Value can be a code meaning (str)
        or a DICOM dataset containing the code.

        Parameters
        ----------
        value: Dataset
            The dataset for creating the code.

        Returns
        -------
        List[ConceptCodeType]
            Concept codes created from dataset.

        """
        codes = cls._from_ds(ds)
        if codes is None:
            return []
        return codes


class CidConceptCode(ConceptCode):
    """Code for concepts  # type: ignore defined in Context groups"""

    cid: ClassVar[Dict[str, Code]]

    def __init__(
        self,
        meaning: str,
        value: Optional[str] = None,
        scheme_designator: Optional[str] = None,
        scheme_version: Optional[str] = None,
    ):
        if value is None or scheme_designator is None:
            code = self._from_meaning(meaning)
            super().__init__(
                value=code.value,
                scheme_designator=code.scheme_designator,
                meaning=code.meaning,
                scheme_version=code.scheme_version,
            )
        else:
            super().__init__(
                value=value,
                scheme_designator=scheme_designator,
                meaning=meaning,
                scheme_version=scheme_version,
            )

    @classmethod
    def _from_meaning(cls: Type[CidConceptCodeType], meaning: str) -> Code:
        try:
            return next(
                code
                for code in cls.cid.values()
                if code.meaning.lower() == meaning.lower()
            )
        except StopIteration:
            raise ValueError(
                f"There is no code with meaning '{meaning}' in {list(cls.cid.keys())}."
            )

    @classmethod
    def list(cls) -> List[str]:
        """Return possible meanings for concept.

        Returns
        -------
        List[str]
            Possible meanings for concept.

        """
        return [code.meaning for code in cls.cid.values()]


class UnitCode(SingleConceptCode):
    """Code for concepts representing units according to UCUM scheme"""

    sequence_name = "MeasurementUnitsCodeSequence"

    def __init__(
        self,
        meaning: str,
        value: Optional[str] = None,
        scheme_designator: Optional[str] = None,
        scheme_version: Optional[str] = None,
    ):
        if value is None or scheme_designator is None:
            code = self._from_unit(meaning)
            super().__init__(
                value=code.value,
                scheme_designator=code.scheme_designator,
                meaning=code.meaning,
                scheme_version=code.scheme_version,
            )
        else:
            super().__init__(
                value=value,
                scheme_designator=scheme_designator,
                meaning=meaning,
                scheme_version=scheme_version,
            )

    @classmethod
    def from_unit(cls, unit: str) -> "UnitCode":
        """Return UCUM scheme ConceptCode.

        Parameters
        ----------
        meaning: str
            Code meaning.

        Returns
        -------
        ConceptCode
            Code created from meaning.

        """
        code = cls._from_unit(unit)
        return cls(
            value=code.value,
            scheme_designator=code.scheme_designator,
            meaning=code.meaning,
            scheme_version=code.scheme_version,
        )

    @classmethod
    def _from_unit(cls, unit: str) -> Code:
        """Return UCUM scheme ConceptCode.

        Parameters
        ----------
        meaning: str
            Code meaning.

        Returns
        ----------
        ConceptCode
            Code created from meaning.

        """
        if unit == "1":
            meaning = "unary"
        elif unit == "{ratio}":
            meaning = "ratio"
        else:
            meaning = unit
        return Code(value=unit, scheme_designator="UCUM", meaning=meaning)

    @classmethod
    def meanings(cls) -> List[str]:
        """Return possible meanings for concept.

        Returns
        -------
        List[str]
            Possible meanings for concept.

        """
        return []


class MeasurementCode(CidConceptCode, SingleConceptCode):
    """
    Concept code for measurement type.
    Microscopy Measurement Types
    """

    sequence_name = "ConceptNameCodeSequence"
    cid = {"Area": Code("42798000", "SCT", "Area")}


class AnnotationTypeCode(CidConceptCode, SingleConceptCode):
    """
    Concept code for annotation type.
    Microscopy Annotation Property Types
    """

    sequence_name = "AnnotationPropertyTypeCodeSequence"
    cid = {
        "Nucleus": Code("84640000", "SCT", "Nucleus"),
        "EntireCell": Code("4421005", "SCT", "Cell"),
    }


class AnnotationCategoryCode(CidConceptCode, SingleConceptCode):
    """
    Concept code for annotation category.
    From CID 7150 Segmentation Property Categories
    """

    sequence_name = "AnnotationPropertyCategoryCodeSequence"
    cid = codes.cid7150.concepts  # type: ignore


class IlluminationCode(CidConceptCode, MultipleConceptCode):
    """
    Concept code for illumination type.
    From CID 8123 Microscopy Illumination Method
    """

    sequence_name = "IlluminationTypeCodeSequence"
    cid = codes.cid8123.concepts  # type: ignore


class LenseCode(CidConceptCode, MultipleConceptCode):
    """
    Concept code for lense.
    From CID 8121 Microscopy Lens Type
    """

    sequence_name = "LensesCodeSequence"
    cid = codes.cid8121.concepts  # type: ignore


class LightPathFilterCode(CidConceptCode, MultipleConceptCode):
    """
    Concept code for light path filter.
    From CID 8124 Microscopy Filter
    """

    sequence_name = "LightPathFilterTypeStackCodeSequence"
    cid = codes.cid8124.concepts  # type: ignore


class ImagePathFilterCode(CidConceptCode, MultipleConceptCode):
    """
    Concept code for image path filter.
    From CID 8124 Microscopy Filter
    """

    sequence_name = "ImagePathFilterTypeStackCodeSequence"
    cid = codes.cid8124.concepts  # type: ignore


class IlluminationColorCode(CidConceptCode, SingleConceptCode):
    """
    Concept code for illumination color.
    From CID 8122 Microscopy Illuminator and Sensor Color
    """

    sequence_name = "IlluminationColorCodeSequence"
    cid = codes.cid8122.concepts  # type: ignore


class IlluminatorCode(CidConceptCode, SingleConceptCode):
    """
    Concept code for illuminator type.
    From CID 8125 Microscopy Illuminator Type
    """

    sequence_name = "IlluminatorTypeCodeSequence"
    cid = codes.cid8125.concepts  # type: ignore


class ChannelDescriptionCode(CidConceptCode, MultipleConceptCode):
    """
    Concept code for channel descriptor.
    From CID 8122 Microscopy Illuminator and Sensor Color
    """

    sequence_name = "ChannelDescriptionCodeSequence"
    cid = codes.cid8122.concepts  # type: ignore


class SpecimenCollectionProcedureCode(CidConceptCode, SingleConceptCode):
    sequence_name = "ConceptCodeSequence"
    cid = codes.cid8109.concepts  # type: ignore


class SpecimenSamplingProcedureCode(CidConceptCode, SingleConceptCode):
    sequence_name = "ConceptCodeSequence"
    cid = codes.cid8110.concepts  # type: ignore


class SpecimenPreparationProcedureCode(CidConceptCode, SingleConceptCode):
    sequence_name = "ConceptCodeSequence"
    cid = codes.cid8111.concepts  # type: ignore


class SpecimenStainsCode(CidConceptCode, SingleConceptCode):
    sequence_name = "ConceptCodeSequence"
    cid = codes.cid8112.concepts  # type: ignore


class SpecimenPreparationStepsCode(CidConceptCode, SingleConceptCode):
    sequence_name = "ConceptCodeSequence"
    cid = codes.cid8113.concepts  # type: ignore


class SpecimenFixativesCode(CidConceptCode, SingleConceptCode):
    sequence_name = "ConceptCodeSequence"
    cid = codes.cid8114.concepts  # type: ignore


class SpecimenEmbeddingMediaCode(CidConceptCode, SingleConceptCode):
    sequence_name = "ConceptCodeSequence"
    cid = codes.cid8115.concepts  # type: ignore


class AnatomicPathologySpecimenTypesCode(CidConceptCode, SingleConceptCode):
    sequence_name = "ConceptCodeSequence"
    cid = codes.cid8103.concepts  # type: ignore


class ContainerComponentTypeCode(CidConceptCode, SingleConceptCode):
    sequence_name = "ConceptCodeSequence"
    cid = codes.cid8102.concepts  # type: ignore


class ContainerTypeCode(CidConceptCode, SingleConceptCode):
    sequence_name = "ConceptCodeSequence"
    cid = codes.cid8101.concepts  # type: ignore


class ConceptNameCode(SingleConceptCode):
    sequence_name = "ConceptNameCodeSequence"

    @classmethod
    def list(cls) -> List[str]:
        return []


def dataset_to_code(dataset: Dataset) -> Code:
    version = dataset.get("CodingSchemeVersion", None)
    if version == "":
        version = None
    return Code(
        dataset.CodeValue,
        dataset.CodingSchemeDesignator,
        dataset.CodeMeaning,
        version,
    )



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import io
import logging
from dataclasses import dataclass
from enum import Enum, IntEnum, auto
from functools import cached_property
from typing import Any, List, Optional, Sequence, Tuple, Union, cast

from pydicom import dcmread
from pydicom.dataset import Dataset
from pydicom.multival import MultiValue
from pydicom.sequence import Sequence as DicomSequence
from pydicom.tag import BaseTag
from pydicom.uid import (
    UID,
    generate_uid,
)
from pydicom.valuerep import DSfloat

from wsidicom.codec import LossyCompressionIsoStandard
from wsidicom.config import settings
from wsidicom.errors import (
    WsiDicomError,
    WsiDicomFileError,
    WsiDicomRequirementError,
    WsiDicomStrictRequirementError,
    WsiDicomUidDuplicateError,
)
from wsidicom.geometry import Size, SizeMm
from wsidicom.instance.image_data import ImageData
from wsidicom.uid import WSI_SOP_CLASS_UID, FileUids, SlideUids


class ImageType(Enum):
    VOLUME = "VOLUME"
    LABEL = "LABEL"
    OVERVIEW = "OVERVIEW"


class Requirement(IntEnum):
    ALWAYS = auto()  # Always required, even if not in strict mode
    STRICT = auto()  # Required if in strict mode
    STANDARD = auto()  # Required or optional in standard, not (yet) needed


@dataclass
class WsiAttributeRequirement:
    """Class for defining requirement and optionally default value for DICOM
    WSI attribute. If image types is given, requirement is only checked for
    images of matching image type.
    """

    requirement: Requirement
    image_types: Tuple[ImageType, ...]
    default: Any = None

    def __init__(
        self,
        requirement: Requirement,
        image_type: Optional[Union[ImageType, Sequence[ImageType]]] = None,
        default: Any = None,
    ) -> None:
        self.requirement = requirement
        if image_type is not None:
            if not isinstance(image_type, Sequence):
                self.image_types = (image_type,)
            else:
                self.image_types = tuple(image_type)
        else:
            self.image_types = ()
        self.default = default

    def get_default(self, image_type: ImageType) -> Any:
        """Get default value for attribute. Raises WsiDicomRequirementError if
        attribute is set as always required for image type. Raises
        WsiDicomStrictRequirementError if attribute is set as always required
        in strict mode."""
        if self.requirement == Requirement.ALWAYS and image_type in self.image_types:
            raise WsiDicomRequirementError(
                f"Attribute is set as always required for {image_type}"
            )
        elif (
            settings.strict_attribute_check
            and self.requirement == Requirement.STRICT
            and image_type in self.image_types
        ):
            raise WsiDicomStrictRequirementError(
                f"Attribute is set as required for {image_type} " "and mode is strict"
            )
        return self.default

    def evaluate(self, image_type: ImageType) -> bool:
        """Evaluate if attribute is required for image type."""
        if self.requirement == Requirement.ALWAYS or (
            self.requirement == Requirement.STRICT and settings.strict_attribute_check
        ):
            if image_type in self.image_types:
                return True
        return False


WSI_ATTRIBUTES = {
    "SOPClassUID": WsiAttributeRequirement(Requirement.ALWAYS),
    "SOPInstanceUID": WsiAttributeRequirement(Requirement.ALWAYS),
    "StudyInstanceUID": WsiAttributeRequirement(Requirement.ALWAYS),
    "SeriesInstanceUID": WsiAttributeRequirement(Requirement.ALWAYS),
    "Rows": WsiAttributeRequirement(Requirement.ALWAYS),
    "Columns": WsiAttributeRequirement(Requirement.ALWAYS),
    "SamplesPerPixel": WsiAttributeRequirement(Requirement.ALWAYS),
    "PhotometricInterpretation": WsiAttributeRequirement(Requirement.ALWAYS),
    "TotalPixelMatrixColumns": WsiAttributeRequirement(Requirement.ALWAYS),
    "TotalPixelMatrixRows": WsiAttributeRequirement(Requirement.ALWAYS),
    "ImageType": WsiAttributeRequirement(Requirement.ALWAYS),
    "SharedFunctionalGroupsSequence": WsiAttributeRequirement(
        Requirement.STRICT, ImageType.VOLUME
    ),
    "FrameOfReferenceUID": WsiAttributeRequirement(Requirement.STRICT),
    "FocusMethod": WsiAttributeRequirement(
        Requirement.STRICT, ImageType.VOLUME, "AUTO"
    ),
    "ExtendedDepthOfField": WsiAttributeRequirement(
        Requirement.STRICT, ImageType.VOLUME, "NO"
    ),
    "OpticalPathSequence": WsiAttributeRequirement(Requirement.STRICT),
    "ImagedVolumeWidth": WsiAttributeRequirement(Requirement.STRICT, ImageType.VOLUME),
    "ImagedVolumeHeight": WsiAttributeRequirement(Requirement.STRICT, ImageType.VOLUME),
    "ImagedVolumeDepth": WsiAttributeRequirement(Requirement.STRICT, ImageType.VOLUME),
    "TotalPixelMatrixFocalPlanes": WsiAttributeRequirement(
        Requirement.STANDARD, default=1
    ),
    "NumberOfOpticalPaths": WsiAttributeRequirement(Requirement.STANDARD, default=1),
    "NumberOfFrames": WsiAttributeRequirement(Requirement.STANDARD, default=1),
    "Modality": WsiAttributeRequirement(Requirement.STANDARD, default="SM"),
    "Manufacturer": WsiAttributeRequirement(Requirement.STANDARD),
    "ManufacturerModelName": WsiAttributeRequirement(Requirement.STANDARD),
    "DeviceSerialNumber": WsiAttributeRequirement(Requirement.STANDARD),
    "SoftwareVersions": WsiAttributeRequirement(Requirement.STANDARD),
    "BitsAllocated": WsiAttributeRequirement(Requirement.STANDARD),
    "BitsStored": WsiAttributeRequirement(Requirement.STANDARD),
    "HighBit": WsiAttributeRequirement(Requirement.STANDARD),
    "PixelRepresentation": WsiAttributeRequirement(Requirement.STANDARD),
    "TotalPixelMatrixOriginSequence": WsiAttributeRequirement(Requirement.STANDARD),
    "ImageOrientationSlide": WsiAttributeRequirement(
        Requirement.STANDARD, default=[0, 1, 0, 1, 0, 0]
    ),
    "AcquisitionDateTime": WsiAttributeRequirement(Requirement.STANDARD),
    "LossyImageCompression": WsiAttributeRequirement(Requirement.STANDARD),
    "VolumetricProperties": WsiAttributeRequirement(
        Requirement.STANDARD, default="VOLUME"
    ),
    "SpecimenLabelInImage": WsiAttributeRequirement(Requirement.STANDARD, default="NO"),
    "BurnedInAnnotation": WsiAttributeRequirement(Requirement.STANDARD, default="NO"),
    "ContentDate": WsiAttributeRequirement(Requirement.STANDARD),
    "ContentTime": WsiAttributeRequirement(Requirement.STANDARD),
    "InstanceNumber": WsiAttributeRequirement(Requirement.STANDARD),
    "DimensionOrganizationSequence": WsiAttributeRequirement(Requirement.STANDARD),
    "ContainerIdentifier": WsiAttributeRequirement(Requirement.STANDARD),
    "SpecimenDescriptionSequence": WsiAttributeRequirement(Requirement.STANDARD),
    "SpacingBetweenSlices": WsiAttributeRequirement(Requirement.STANDARD, default=0.0),
    "SOPInstanceUIDOfConcatenationSource": WsiAttributeRequirement(
        Requirement.STANDARD
    ),
    "DimensionOrganizationType": WsiAttributeRequirement(
        Requirement.STANDARD, default="TILED_SPARSE"
    ),
    "NumberOfFocalPlanes": WsiAttributeRequirement(Requirement.STANDARD, default=1),
    "DistanceBetweenFocalPlanes": WsiAttributeRequirement(
        Requirement.STANDARD, default=0.0
    ),
    "SliceThickness": WsiAttributeRequirement(Requirement.STANDARD),
}


class TileType(Enum):
    FULL = "TILED_FULL"
    SPARSE = "TILED_SPARSE"


class WsiDataset(Dataset):
    """Extend pydicom.dataset.Dataset (containing WSI metadata) with simple
    parsers for attributes specific for WSI. Use snake case to avoid name
    collision with dicom fields (that are handled by pydicom.dataset.Dataset).
    """

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self})"

    def __str__(self) -> str:
        return f"{type(self).__name__} of dataset {self.uids.instance}"

    @cached_property
    def uids(self) -> FileUids:
        """Return UIDs from dataset.

        Returns
        -------
        FileUids
            Found UIDs from dataset.
        """
        instance_uid = UID(self.SOPInstanceUID)
        concatenation_uid = self._get_dicom_attribute(
            "SOPInstanceUIDOfConcatenationSource"
        )
        frame_of_reference_uid = self._get_dicom_attribute("FrameOfReferenceUID")

        slide_uids = SlideUids(
            self.StudyInstanceUID,
            self.SeriesInstanceUID,
            frame_of_reference_uid,
        )
        file_uids = FileUids(instance_uid, concatenation_uid, slide_uids)
        return file_uids

    @cached_property
    def frame_offset(self) -> int:
        """Return frame offset (offset to first frame in instance if
        concatenated). Is zero if non-concatenated instance or first instance
        in concatenated instance.

        Returns
        -------
        int
            Concatenation offset in number of frames.
        """
        if self.uids.concatenation is None:
            return 0
        try:
            return int(self.ConcatenationFrameOffsetNumber)
        except AttributeError:
            raise WsiDicomError(
                "Concatenated file missing concatenation frame offset" "number"
            )

    @property
    def frame_count(self) -> int:
        """Return number of frames in instance."""
        frame_count = self._get_dicom_attribute("NumberOfFrames")
        return cast(int, frame_count)

    @cached_property
    def tile_type(self) -> TileType:
        """Return tiling type of dataset. Raises WsiDicomError if type
        is undetermined.

        Returns
        -------
        TileType
            Tiling type
        """
        tile_type = self._get_dicom_attribute("DimensionOrganizationType")
        if tile_type == "TILED_FULL":
            # By the standard it should be tiled full.
            return TileType.FULL
        if "PerFrameFunctionalGroupsSequence" in self:
            # If no per frame functional sequence we can't make a sparse tile index.
            return TileType.SPARSE
        if self.image_type == ImageType.LABEL:
            # Labels are expected to only have one frame and can be treated as tiled full.
            return TileType.FULL
        number_of_focal_planes = getattr(self, "TotalPixelMatrixFocalPlanes", 1)
        number_of_optical_paths = getattr(self, "NumberOfOpticalPaths", 1)
        if self.frame_count == number_of_focal_planes * number_of_optical_paths:
            # One frame per focal plane and optical path, treat as tiled full.
            return TileType.FULL
        raise WsiDicomError("Undetermined tile type.")

    @cached_property
    def pixel_measure(self) -> Optional[Dataset]:
        """Return Pixel measure dataset from dataset if found.

        Returns
        -------
        Optional[Dataset]
            Found Pixel measure dataset.
        """
        shared_functional_group = self._get_dicom_attribute(
            "SharedFunctionalGroupsSequence"
        )
        if shared_functional_group is None:
            return None
        pixel_measure_sequence = self._get_dicom_attribute(
            "PixelMeasuresSequence", shared_functional_group[0]
        )
        if pixel_measure_sequence is None:
            return None
        return pixel_measure_sequence[0]

    @cached_property
    def pixel_spacing(self) -> Optional[SizeMm]:
        """Read pixel spacing from dicom dataset.

        Parameters
        ----------
        ds: Dataset
            Pydicom dataset

        Returns
        -------
        SizeMm
            The pixel spacing in mm/pixel.
        """
        if self.pixel_measure is None:
            return None
        pixel_spacing_values = getattr(self.pixel_measure, "PixelSpacing", None)
        if pixel_spacing_values is not None:
            if any([spacing <= 0 for spacing in pixel_spacing_values]):
                logging.warning(f"Pixel spacing not positive, {pixel_spacing_values}")
                return None
            return SizeMm(pixel_spacing_values[1], pixel_spacing_values[0])
        return None

    @cached_property
    def spacing_between_slices(self) -> Optional[float]:
        """Return spacing between slices."""
        if self.pixel_measure is None:
            return None
        return getattr(self.pixel_measure, "SpacingBetweenSlices", None)

    @cached_property
    def frame_sequence(self) -> DicomSequence:
        """Return per frame functional group sequence if present, otherwise
        shared functional group sequence.

        Returns
        -------
        DicomSequence
            Per frame or shared functional group sequence.
        """
        if "PerFrameFunctionalGroupsSequence" in self and (
            "PlanePositionSlideSequence" in self.PerFrameFunctionalGroupsSequence[0]
        ):
            return self.PerFrameFunctionalGroupsSequence
        elif "SharedFunctionalGroupsSequence" in self:
            return self.SharedFunctionalGroupsSequence
        return DicomSequence([])

    @property
    def ext_depth_of_field(self) -> bool:
        """Return true if instance has extended depth of field
        (several focal planes are combined to one plane)."""
        return self._ext_depth_of_field[0]

    @property
    def ext_depth_of_field_planes(self) -> Optional[int]:
        """Return number of focal planes used for extended depth of
        field."""
        return self._ext_depth_of_field[1]

    @property
    def ext_depth_of_field_plane_distance(self) -> Optional[float]:
        """Return total focal depth used for extended depth of field."""
        return self._ext_depth_of_field[0]

    @cached_property
    def focus_method(self) -> str:
        """Return focus method."""
        focus_method = self._get_dicom_attribute("FocusMethod")
        return str(focus_method)

    @cached_property
    def image_size(self) -> Size:
        """Read total pixel size from dataset.

        Returns
        -------
        Size
            The image size
        """
        image_size = Size(self.TotalPixelMatrixColumns, self.TotalPixelMatrixRows)
        if image_size.width <= 0 or image_size.height <= 0:
            raise WsiDicomError("Image size is zero")
        if self.tile_type == TileType.FULL and self.uids.concatenation is None:
            # Check that the number of frames match the image size and tile size.
            # Dont check concatenated instances as the frame count is ambiguous.
            expected_tiled_size = image_size.ceil_div(self.tile_size)
            number_of_focal_planes = getattr(self, "TotalPixelMatrixFocalPlanes", 1)
            number_of_optical_paths = getattr(self, "NumberOfOpticalPaths", 1)
            expected_frame_count = (
                expected_tiled_size.area
                * number_of_focal_planes
                * number_of_optical_paths
            )
            if expected_frame_count != self.frame_count:
                error = (
                    f"Image size {image_size} does not match tile size "
                    f"{self.tile_size} and number of frames {self.frame_count} "
                    f"for tile type {TileType.FULL}."
                )
                if (
                    self.image_type == ImageType.VOLUME
                    and self.frame_count
                    != number_of_focal_planes * number_of_optical_paths
                ):
                    # Be strict on volume images if more than one frame per focal plane
                    # and optical path.
                    raise WsiDicomError(error)
                # Labels and overviews are likely to have only one tile.
                error += " Overriding image size to tile size."
                logging.warning(error)
                image_size = self.tile_size
        return image_size

    @cached_property
    def mm_size(self) -> Optional[SizeMm]:
        """Read mm size from dataset.

        Returns
        -------
        SizeMm
            The size of the image in mm
        """
        mm_width = self._get_dicom_attribute("ImagedVolumeWidth")
        mm_height = self._get_dicom_attribute("ImagedVolumeHeight")
        if mm_width is None or mm_height is None:
            mm_size = None
        else:
            mm_size = SizeMm(mm_width, mm_height)
        return mm_size

    @cached_property
    def mm_depth(self) -> Optional[float]:
        """Return depth of image in mm."""
        return self._get_dicom_attribute("ImagedVolumeDepth")

    @cached_property
    def tile_size(self) -> Size:
        """Read tile size from from dataset.

        Returns
        -------
        Size
            The tile size
        """
        return Size(self.Columns, self.Rows)

    @property
    def samples_per_pixel(self) -> int:
        """Return samples per pixel (3 for RGB)."""
        return self.SamplesPerPixel

    @property
    def bits(self) -> int:
        """Return the number of bits stored for each sample."""
        return self.BitsStored

    @property
    def lossy_compressed(self) -> bool:
        """Return true if image has been lossy compressed."""
        return self._get_dicom_attribute("LossyImageCompression") == "01"

    @property
    def photometric_interpretation(self) -> str:
        """Return photometric interpretation."""
        return self.PhotometricInterpretation

    @cached_property
    def optical_path_sequence(self) -> Optional[DicomSequence]:
        """Return optical path sequence from dataset."""
        return self._get_dicom_attribute("OpticalPathSequence")

    @property
    def slice_thickness(self) -> Optional[float]:
        """Return slice thickness spacing from pixel measure dataset.

        Returns
        -------
        Optional[float]
            Slice thickess or None if unknown.
        """
        try:
            return self._get_dicom_attribute("SliceThickness", self.pixel_measure)
        except AttributeError:
            if self.mm_depth is not None:
                number_of_focal_planes = getattr(self, "TotalPixelMatrixFocalPlanes", 1)
                return self.mm_depth / number_of_focal_planes
        return None

    @cached_property
    def image_type(self) -> ImageType:
        """Return wsi flavour from wsi type tuple.

        Returns
        -------
        ImageType
            Wsi flavour.
        """
        return self._get_image_type(self.ImageType)

    @classmethod
    def is_supported_wsi_dicom(cls, dataset: Dataset) -> Optional[ImageType]:
        """Check if dataset is dicom wsi type and that required attributes
        (for the function of the library) is available.
        Warn if attribute listed as required in the library or required in the
        standard is missing.

        Parameters
        ----------
        dataset: Dataset
            Pydicom dataset to check if is a WSI dataset.

        Returns
        -------
        Optional[ImageType]
            WSI image flavor
        """

        sop_class_uid: UID = getattr(dataset, "SOPClassUID")
        if sop_class_uid != WSI_SOP_CLASS_UID:
            logging.debug(f"Non-wsi image, SOP class {sop_class_uid.name}.")
            return None

        try:
            image_type = cls._get_image_type(dataset.ImageType)
        except ValueError:
            logging.debug(f"Non-supported image type {dataset.ImageType}.")
            return None

        for name, attribute in WSI_ATTRIBUTES.items():
            if name not in dataset and attribute.evaluate(image_type):
                logging.debug(f"Missing required attribute {name}.")
                return None
        pixel_represention = int(getattr(dataset, "PixelRepresentation", 0))
        if pixel_represention != 0:
            logging.debug(f"Unsupported pixel representation {pixel_represention}.")
            return None
        planar_configuration = int(getattr(dataset, "PlanarConfiguration", 0))
        if planar_configuration != 0:
            logging.debug(f"Unsupported planar configuration {planar_configuration}.")
            return None
        return image_type

    @staticmethod
    def check_duplicate_dataset(
        datasets: Sequence["WsiDataset"], caller: object
    ) -> None:
        """Check for duplicates in a list of datasets. Datasets are duplicate
        if instance uids match. Stops at first found duplicate and raises
        WsiDicomUidDuplicateError.

        Parameters
        ----------
        datasets: Sequence[Dataset]
            List of datasets to check.
        caller: Object
            Object that the files belongs to.
        """
        instance_uids: List[UID] = []

        for dataset in datasets:
            instance_uid = UID(dataset.SOPInstanceUID)
            if instance_uid not in instance_uids:
                instance_uids.append(instance_uid)
            else:
                raise WsiDicomUidDuplicateError(str(dataset), str(caller))

    def matches_instance(self, other_dataset: "WsiDataset") -> bool:
        """Return true if other file is of the same instance as self.

        Parameters
        ----------
        other_dataset: 'WsiDataset'
            Dataset to check.

        Returns
        -------
        bool
            True if same instance.
        """

        return (
            self.uids == other_dataset.uids
            and self.image_size == other_dataset.image_size
            and self.tile_size == other_dataset.tile_size
            and self.tile_type == other_dataset.tile_type
            and (
                getattr(self, "TotalPixelMatrixOriginSequence", None)
                == getattr(other_dataset, "TotalPixelMatrixOriginSequence", None)
            )
        )

    def matches_series(self, uids: SlideUids, tile_size: Optional[Size] = None) -> bool:
        """Check if instance is valid (Uids and tile size match).
        Base uids should match for instances in all types of series,
        tile size should only match for level series.
        """
        if tile_size is not None and tile_size != self.tile_size:
            return False

        return self.uids.slide.matches(uids)

    def read_optical_path_identifier(self, frame: Dataset) -> str:
        """Return optical path identifier from frame, or from self if not
        found."""
        optical_sequence = getattr(
            frame, "OpticalPathIdentificationSequence", self.optical_path_sequence
        )
        if optical_sequence is None:
            return "0"
        return getattr(optical_sequence[0], "OpticalPathIdentifier", "0")

    def get_multi_value(self, tag: Union[str, BaseTag]) -> List[Any]:
        """Return values for tag as list of values. If tag is not found, return empty
        list. If tag is not multi value, return list with one value.

        Parameters
        ----------
        tag: Union[str, BaseTag]
            Tag to get values for.

        Returns
        -------
        List[Any]
            List of values.
        """
        element = self.get(tag)
        if element is None:
            return []
        vm = getattr(element, "VM", 1)
        if vm > 1 or isinstance(element, MultiValue):
            return [value for value in element]
        return [element.value]

    def as_tiled_full(
        self,
        focal_planes: Sequence[float],
        optical_paths: Sequence[str],
        tiled_size: Size,
        scale: int = 1,
    ) -> "WsiDataset":
        """Return copy of dataset with properties set to reflect a tiled full
        arrangement of the listed image data. Optionally set properties to
        reflect scaled data.

        Parameters
        ----------
        focal_planes: Sequence[float]
            Focal planes that should be encoded into dataset.
        optical_paths: Sequence[str]
            Optical paths that should be encoded into dataset.
        tiled_size: Size
            Tiled size of image.
        scale: int = 1
            Optionally scale data.

        Returns
        -------
        WsiDataset
            Copy of dataset set as tiled full.

        """
        # Workaround for deepcopy not working
        # See https://github.com/pydicom/pydicom/issues/1816
        with io.BytesIO() as buffer:
            self.is_implicit_VR = False
            self.is_little_endian = True
            self.save_as(buffer)
            buffer.seek(0)
            dataset = WsiDataset(dcmread(buffer, force=True))
        dataset.DimensionOrganizationType = "TILED_FULL"
        # Make a new Shared functional group sequence and Pixel measure
        # sequence if not in dataset, otherwise update the Pixel measure
        # sequence
        shared_functional_group = getattr(
            dataset, "SharedFunctionalGroupsSequence", DicomSequence([Dataset()])
        )

        plane_position_slide = Dataset()
        plane_position_slide.ZOffsetInSlideCoordinateSystem = DSfloat(
            focal_planes[0], True
        )
        shared_functional_group[0].PlanePositionSlideSequence = DicomSequence(
            [plane_position_slide]
        )

        pixel_measure = getattr(
            shared_functional_group[0],
            "PixelMeasuresSequence",
            DicomSequence([Dataset()]),
        )
        if dataset.pixel_spacing is not None:
            pixel_measure[0].PixelSpacing = [
                DSfloat(dataset.pixel_spacing.height * scale, True),
                DSfloat(dataset.pixel_spacing.width * scale, True),
            ]
        pixel_measure[0].SpacingBetweenSlices = DSfloat(
            self._get_spacing_between_slices_for_focal_planes(focal_planes), True
        )

        if self.slice_thickness is not None:
            pixel_measure[0].SliceThickness = DSfloat(dataset.slice_thickness, True)

        shared_functional_group[0].PixelMeasuresSequence = pixel_measure
        dataset.SharedFunctionalGroupsSequence = shared_functional_group

        # Remove Per Frame functional groups sequence
        if "PerFrameFunctionalGroupsSequence" in dataset:
            del dataset["PerFrameFunctionalGroupsSequence"]

        dataset.TotalPixelMatrixFocalPlanes = len(focal_planes)
        dataset.NumberOfOpticalPaths = len(optical_paths)
        dataset.NumberOfFrames = (
            max(tiled_size.ceil_div(scale).area, 1)
            * len(focal_planes)
            * len(optical_paths)
        )
        scaled_size = dataset.image_size.ceil_div(scale)
        dataset.TotalPixelMatrixColumns = max(scaled_size.width, 1)
        dataset.TotalPixelMatrixRows = max(scaled_size.height, 1)
        return dataset

    @classmethod
    def create_instance_dataset(
        cls,
        dataset: Dataset,
        image_type: ImageType,
        image_data: ImageData,
        pyramid_index: Optional[int] = None,
    ) -> "WsiDataset":
        """Return instance dataset for image_data based on base dataset.

        Parameters
        ----------
        base_dataset: Dataset
            Dataset common for all instances.
        image_type:
            Type of instance ('VOLUME', 'LABEL', 'OVERVIEW)
        image_data:
            Image data to create dataset for.

        Returns
        -------
        WsiDataset
            Dataset for instance.
        """
        if image_type == ImageType.VOLUME and pyramid_index == 0:
            resampled = "NONE"
        else:
            resampled = "RESAMPLED"
        dataset.ImageType = ["ORIGINAL", "PRIMARY", image_type.value, resampled]
        dataset.SOPInstanceUID = generate_uid(prefix=None)
        shared_functional_group_sequence = Dataset()
        if image_data.pixel_spacing is None:
            if image_type == ImageType.VOLUME:
                raise ValueError(
                    "Image flavor 'VOLUME' requires pixel spacing to be set"
                )
        else:
            pixel_measure_sequence = Dataset()
            pixel_measure_sequence.PixelSpacing = [
                DSfloat(image_data.pixel_spacing.height, True),
                DSfloat(image_data.pixel_spacing.width, True),
            ]
            pixel_measure_sequence.SpacingBetweenSlices = DSfloat(0.0, True)
            # DICOM 2022a part 3 IODs - C.8.12.4.1.2 Imaged Volume Width,
            # Height, Depth. Depth must not be 0. Default to 0.5 microns
            slice_thickness = 0.0005
            pixel_measure_sequence.SliceThickness = DSfloat(0.0005, True)
            shared_functional_group_sequence.PixelMeasuresSequence = DicomSequence(
                [pixel_measure_sequence]
            )
            dataset.SharedFunctionalGroupsSequence = DicomSequence(
                [shared_functional_group_sequence]
            )
            dataset.ImagedVolumeWidth = (
                image_data.image_size.width * image_data.pixel_spacing.width
            )
            dataset.ImagedVolumeHeight = (
                image_data.image_size.height * image_data.pixel_spacing.height
            )
            # SliceThickness is in mm, ImagedVolumeDepth in um
            dataset.ImagedVolumeDepth = DSfloat(slice_thickness * 1000, True)
            # DICOM 2022a part 3 IODs - C.8.12.9 Whole Slide Microscopy Image
            # Frame Type Macro. Analogous to ImageType and shared by all
            # frames so clone
            wsi_frame_type_item = Dataset()
            wsi_frame_type_item.FrameType = dataset.ImageType
            (
                shared_functional_group_sequence.WholeSlideMicroscopyImageFrameTypeSequence
            ) = DicomSequence([wsi_frame_type_item])

        if image_data.image_coordinate_system is not None:
            dataset.ImageOrientationSlide = [
                DSfloat(value, True)
                for value in image_data.image_coordinate_system.orientation.values
            ]
            offset_item = Dataset()
            offset_item.XOffsetInSlideCoordinateSystem = DSfloat(
                image_data.image_coordinate_system.origin.x, True
            )
            offset_item.YOffsetInSlideCoordinateSystem = DSfloat(
                image_data.image_coordinate_system.origin.y, True
            )
            dataset.TotalPixelMatrixOriginSequence = DicomSequence([offset_item])

        dataset.DimensionOrganizationType = "TILED_FULL"
        dataset.TotalPixelMatrixColumns = image_data.image_size.width
        dataset.TotalPixelMatrixRows = image_data.image_size.height
        dataset.Columns = image_data.tile_size.width
        dataset.Rows = image_data.tile_size.height
        dataset.NumberOfFrames = (
            image_data.tiled_size.width * image_data.tiled_size.height
        )
        dataset.BitsAllocated = image_data.bits // 8 * 8
        dataset.BitsStored = image_data.bits
        dataset.HighBit = image_data.bits - 1
        dataset.PixelRepresentation = 0
        if image_data.lossy_compressed:
            method = LossyCompressionIsoStandard.transfer_syntax_to_iso(
                image_data.transfer_syntax
            )
            if method is None:
                raise NotImplementedError(
                    "Creating lossy compressed image with transfer syntax that is not "
                    "lossy is not implemented."
                )
            dataset.LossyImageCompression = "01"
            dataset.LossyImageCompressionRatio = 1
            dataset.LossyImageCompressionMethod = method.value

        dataset.PhotometricInterpretation = image_data.photometric_interpretation
        dataset.SamplesPerPixel = image_data.samples_per_pixel

        if image_data.samples_per_pixel == 3:
            dataset.PlanarConfiguration = 0

        dataset.FocusMethod = "AUTO"
        dataset.ExtendedDepthOfField = "NO"
        return WsiDataset(dataset)

    def _get_dicom_attribute(
        self, name: str, dataset: Optional[Dataset] = None
    ) -> Optional[Any]:
        """Return value for DICOM attribute by name. Optionally get it from
        another dataset. Returns default value if possible when attribute
        is missing.

        Parameters
        ----------
        name: str
            Name of attribute to get.
        dataset: Optional[Dataset] = None
            Optional other dataset to get attribute from.

        Returns
        -------
        Optional[Any]
            Value of attribute, or None.
        """
        if dataset is None:
            dataset = self
        value = getattr(dataset, name, None)
        if value is None:
            return WSI_ATTRIBUTES[name].get_default(self.image_type)
        return value

    @cached_property
    def _ext_depth_of_field(self) -> Tuple[bool, Optional[int], Optional[float]]:
        """Return extended depth of field (enabled, number of focal planes,
        distance between focal planes) from dataset.

        Returns
        -------
        Tuple[bool, Optional[int], Optional[float]]
            If extended depth of field is used, and if used number of focal
            planes and distance between focal planes.
        """
        if self._get_dicom_attribute("ExtendedDepthOfField") != "YES":
            return False, None, None

        planes = self._get_dicom_attribute("NumberOfFocalPlanes")
        distance = self._get_dicom_attribute("DistanceBetweenFocalPlanes")
        if planes is None or distance is None:
            raise WsiDicomFileError(
                self.filepath,
                "Missing NumberOfFocalPlanes or DistanceBetweenFocalPlanes",
            )
        return True, planes, distance

    @staticmethod
    def _get_spacing_between_slices_for_focal_planes(
        focal_planes: Sequence[float],
    ) -> float:
        """Return spacing between slices in mm for focal planes (defined in
        um). Spacing must be the same between all focal planes for TILED_FULL
        arrangement.

        Parameters
        ----------
        focal_planes: Sequence[float]
            Focal planes to calculate spacing for.

        Returns
        -------
        float
            Spacing between focal planes.

        """
        if len(focal_planes) == 1:
            return 0.0
        spacing: Optional[float] = None
        sorted_focal_planes = sorted(focal_planes)
        for index in range(len(sorted_focal_planes) - 1):
            this_spacing = sorted_focal_planes[index + 1] - sorted_focal_planes[index]
            if spacing is None:
                spacing = this_spacing
            elif abs(spacing - this_spacing) > settings.focal_plane_distance_threshold:
                raise NotImplementedError(
                    "Image data has non-equal spacing between slices: "
                    f"{spacing, this_spacing}, difference threshold: "
                    f"{settings.focal_plane_distance_threshold}, "
                    "not possible to encode as TILED_FULL"
                )
        if spacing is None:
            raise ValueError("Could not calculate spacings.")
        return spacing / 1000.0

    @staticmethod
    def _get_image_type(wsi_type: Tuple[str, str, str, str]) -> ImageType:
        """Return wsi flavour from wsi type tuple.

        Returns
        -------
        str
            Wsi flavour.
        """
        IMAGE_TYPE_INDEX_IN_WSI_TYPE = 2
        return ImageType(wsi_type[IMAGE_TYPE_INDEX_IN_WSI_TYPE])



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from collections import defaultdict
from typing import Dict, Iterable, List, Optional, Set, Tuple

from PIL.Image import Image
from pydicom.uid import UID

from wsidicom.config import settings
from wsidicom.errors import (
    WsiDicomMatchError,
    WsiDicomNoResolutionError,
    WsiDicomNotFoundError,
    WsiDicomOutOfBoundsError,
)
from wsidicom.geometry import Point, Region, RegionMm, Size, SizeMm
from wsidicom.instance import (
    ImageData,
    ImageType,
    WsiDataset,
    WsiInstance,
)
from wsidicom.metadata import ImageCoordinateSystem
from wsidicom.stringprinting import dict_pretty_str
from wsidicom.uid import SlideUids


class Group:
    """Represents a group of instances having the same size, but possibly
    different z coordinate and/or optical path."""

    def __init__(self, instances: Iterable[WsiInstance]):
        """Create a group of WsiInstances. Instances should match in the common
        uids, wsi type, and tile size.

        Parameters
        ----------
        instances: Sequence[WsiInstance]
            Instances to build the group.
        """
        self._instances = {  # key is identifier (Uid)
            instance.identifier: instance for instance in instances
        }
        self._validate_group()

        base_instance = next(iter(self._instances.values()))
        self._image_type = base_instance.image_type
        self._uids = base_instance.uids

        self._size = base_instance.size
        self._pixel_spacing = base_instance.pixel_spacing
        self._default_instance_uid: UID = base_instance.identifier

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.instances.values()})"

    def __str__(self) -> str:
        return self.pretty_str()

    def pretty_str(self, indent: int = 0, depth: Optional[int] = None) -> str:
        string = f"Image: size: {self.size} px, mpp: {self.mpp} um/px"
        if depth is not None:
            depth -= 1
            if depth < 0:
                return string
        string += " Instances: " + dict_pretty_str(self.instances, indent, depth)
        return string

    def __getitem__(self, index: UID) -> WsiInstance:
        return self.instances[index]

    @property
    def uids(self) -> SlideUids:
        """Return uids"""
        return self._uids

    @property
    def image_type(self) -> ImageType:
        """Return wsi type"""
        return self._image_type

    @property
    def size(self) -> Size:
        """Return image size in pixels"""
        return self._size

    @property
    def mpp(self) -> Optional[SizeMm]:
        """Return pixel spacing in um/pixel"""
        if self.pixel_spacing is None:
            return None
        return self.pixel_spacing * 1000.0

    @property
    def pixel_spacing(self) -> Optional[SizeMm]:
        """Return pixel spacing in mm/pixel"""
        return self._pixel_spacing

    @property
    def instances(self) -> Dict[UID, WsiInstance]:
        """Return contained instances"""
        return self._instances

    @property
    def default_instance(self) -> WsiInstance:
        """Return default instance"""
        return self.instances[self._default_instance_uid]

    @property
    def datasets(self) -> List[WsiDataset]:
        """Return contained datasets."""
        instance_datasets = [instance.datasets for instance in self.instances.values()]
        return [dataset for sublist in instance_datasets for dataset in sublist]

    @property
    def optical_paths(self) -> List[str]:
        return list(
            {
                path
                for instance in self.instances.values()
                for path in instance.optical_paths
            }
        )

    @property
    def focal_planes(self) -> List[float]:
        return list(
            {
                focal_plane
                for innstance in self.instances.values()
                for focal_plane in innstance.focal_planes
            }
        )

    @property
    def image_coordinate_system(self) -> Optional[ImageCoordinateSystem]:
        return self.default_instance.image_coordinate_system

    def matches(self, other_group: "Group") -> bool:
        """Check if group matches other group. If strict common Uids should
        match. Wsi type should always match.

        Parameters
        ----------
        other_group: Group
            Other group to match against.

        Returns
        -------
        bool
            True if other group matches.
        """
        return (
            self.uids.matches(other_group.uids)
            and other_group.image_type == self.image_type
        )

    def valid_pixels(self, region: Region) -> bool:
        """Check if pixel region is within the size of the group image size.

        Parameters
        ----------
        region: Region
            Pixel region to check

        Returns
        -------
        bool
            True if pixel position and size is within image
        """
        # Return true if inside pixel plane.
        image_region = Region(Point(0, 0), self.size)
        return region.is_inside(image_region)

    def get_instance(
        self, z: Optional[float] = None, path: Optional[str] = None
    ) -> WsiInstance:
        """Search for instance fulfilling the parameters.
        The behavior when z and/or path is none could be made more
        clear.

        Parameters
        ----------
        z: Optional[float] = None
            Z coordinate of the searched instance
        path: Optional[str] = None
            Optical path of the searched instance

        Returns
        -------
        WsiInstance
            The instance containing selected path and z coordinate
        """
        if z is None and path is None:
            return self.default_instance

        # Sort instances by number of focal planes (prefer simplest instance)
        sorted_instances = sorted(
            list(self._instances.values()), key=lambda i: len(i.focal_planes)
        )
        try:
            if z is None:
                # Select the instance with selected optical path
                return next(
                    instance
                    for instance in sorted_instances
                    if path in instance.optical_paths
                )
            if path is None:
                # Select the instance with selected z
                return next(
                    instance
                    for instance in sorted_instances
                    if z in instance.focal_planes
                )

            # Select by both path and z
            return next(
                instance
                for instance in sorted_instances
                if (z in instance.focal_planes and path in instance.optical_paths)
            )
        except StopIteration as exception:
            raise WsiDicomNotFoundError(
                f"Instance for path: {path}, z: {z}", "group"
            ) from exception

    def get_default_full(self) -> Image:
        """Read full image using default z coordinate and path.

        Returns
        -------
        Image
            Full image of the group.
        """
        instance = self.default_instance
        z = instance.default_z
        path = instance.default_path
        region = Region(position=Point(x=0, y=0), size=self.size)
        image = self.get_region(region, z, path)
        return image

    def get_region(
        self,
        region: Region,
        z: Optional[float] = None,
        path: Optional[str] = None,
        threads: int = 1,
    ) -> Image:
        """Read region defined by pixels.

        Parameters
        ----------
        location: int, int
            Upper left corner of region in pixels
        size: int
            Size of region in pixels
        z: Optional[float] = None
            Z coordinate, optional
        path: Optional[str] = None
            optical path, optional
        threads: int = 1
            Number of threads to use for read.

        Returns
        -------
        Image
            Region as image
        """

        instance = self.get_instance(z, path)
        if z is None:
            z = instance.default_z
        if path is None:
            path = instance.default_path
        image = instance.image_data.stitch_tiles(region, path, z, threads)
        return image

    def get_region_mm(
        self,
        region: RegionMm,
        z: Optional[float] = None,
        path: Optional[str] = None,
        slide_origin: bool = False,
        threads: int = 1,
    ) -> Image:
        """Read region defined by mm.

        Parameters
        ----------
        region: RegionMm
            Region defining upper left corner and size in mm.
        z: Optional[float] = None
            Z coordinate, optional.
        path: Optional[str] = None
            optical path, optional.
        slide_origin: bool = False.
            If to use the slide origin instead of image origin.
        threads: int = 1
            Number of threads to use for read.

        Returns
        -------
        Image
            Region as image
        """
        to_coordinate_system = None
        if slide_origin:
            if self.image_coordinate_system is None:
                raise ValueError(
                    "Can't map to slide region as image coordinate system is not defined."
                )
            to_coordinate_system = self.image_coordinate_system
            region = to_coordinate_system.slide_to_image(region)
        pixel_region = self.mm_to_pixel(region)
        image = self.get_region(pixel_region, z, path, threads)
        if to_coordinate_system:
            image = image.rotate(
                to_coordinate_system.rotation,
                resample=settings.pillow_resampling_filter,
                expand=True,
            )
        return image

    def get_tile(
        self, tile: Point, z: Optional[float] = None, path: Optional[str] = None
    ) -> Image:
        """Return tile at tile coordinate x, y as image.

        Parameters
        ----------
        tile: Point
            Tile x, y coordinate
        z: Optional[float] = None
            Z coordinate
        path: Optional[str] = None
            Optical path

        Returns
        -------
        Image
            The tile as image
        """

        instance = self.get_instance(z, path)
        if z is None:
            z = instance.default_z
        if path is None:
            path = instance.default_path
        return instance.image_data.get_tile(tile, z, path)

    def get_encoded_tile(
        self, tile: Point, z: Optional[float] = None, path: Optional[str] = None
    ) -> bytes:
        """Return tile at tile coordinate x, y as bytes.

        Parameters
        ----------
        tile: Point
            Tile x, y coordinate
        z: Optional[float] = None
            Z coordinate
        path: Optional[str] = None
            Optical path

        Returns
        -------
        bytes
            The tile as bytes
        """
        instance = self.get_instance(z, path)
        if z is None:
            z = instance.default_z
        if path is None:
            path = instance.default_path
        return instance.image_data.get_encoded_tile(tile, z, path)

    def mm_to_pixel(self, region: RegionMm) -> Region:
        """Convert region in mm to pixel region.

        Parameters
        ----------
        region: RegionMm
            Region in mm

        Returns
        -------
        Region
            Region in pixels
        """
        if self.pixel_spacing is None:
            raise WsiDicomNoResolutionError()
        pixel_region = Region(
            position=region.position // self.pixel_spacing,
            size=region.size // self.pixel_spacing,
        )
        if not self.valid_pixels(pixel_region):
            raise WsiDicomOutOfBoundsError(
                f"Region {region}", f"level size {self.size}"
            )
        return pixel_region

    def _validate_group(self):
        """Check that no file or instance in group is duplicate, and if strict
        instances in group matches. Raises WsiDicomMatchError otherwise.
        """
        instances = list(self.instances.values())
        base_instance = instances[0]
        for instance in instances[1:]:
            if not base_instance.matches(instance):
                raise WsiDicomMatchError(str(instance), str(self))

        WsiDataset.check_duplicate_dataset(self.datasets, self)
        WsiInstance.check_duplicate_instance(instances, self)

    @staticmethod
    def _get_frame_information(
        data: Dict[Tuple[str, float], ImageData]
    ) -> Tuple[List[float], List[str], Size]:
        """Return optical_paths, focal planes, and tiled size."""
        focal_planes_by_optical_path: Dict[str, Set[float]] = defaultdict(set)
        all_focal_planes: Set[float] = set()
        tiled_sizes: Set[Size] = set()
        for (optical_path, focal_plane), image_data in data.items():
            focal_planes_by_optical_path[optical_path].add(focal_plane)
            all_focal_planes.add(focal_plane)
            tiled_sizes.add(image_data.tiled_size)

        focal_planes_sparse_by_optical_path = any(
            optical_path_focal_planes != all_focal_planes
            for optical_path_focal_planes in focal_planes_by_optical_path.values()
        )
        if focal_planes_sparse_by_optical_path:
            raise ValueError("Each optical path must have the same focal planes.")

        if len(tiled_sizes) != 1:
            raise ValueError("Expected only one tiled size")
        tiled_size = list(tiled_sizes)[0]
        return (
            sorted(list(all_focal_planes)),
            sorted(list(focal_planes_by_optical_path.keys())),
            tiled_size,
        )



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from abc import ABCMeta, abstractmethod
from functools import cached_property
from typing import Iterator, List, Optional, Sequence

from PIL.Image import Image

from wsidicom.codec import Codec, Decoder
from wsidicom.errors import WsiDicomOutOfBoundsError
from wsidicom.geometry import Point, Region, Size, SizeMm
from wsidicom.instance.dataset import TileType, WsiDataset
from wsidicom.instance.image_data import ImageData
from wsidicom.instance.tile_index.full_tile_index import FullTileIndex
from wsidicom.instance.tile_index.sparse_tile_index import SparseTileIndex
from wsidicom.instance.tile_index.tile_index import TileIndex
from wsidicom.metadata.schema.dicom.image import ImageCoordinateSystemDicomSchema
from wsidicom.metadata.image import ImageCoordinateSystem


class WsiDicomImageData(ImageData, metaclass=ABCMeta):
    def __init__(self, datasets: Sequence[WsiDataset], codec: Codec):
        self._datasets = datasets
        self._decoder = codec.decoder
        super().__init__(codec.encoder)

    @abstractmethod
    def _get_tile_frame(self, frame_index: int) -> bytes:
        """Return tile frame for frame index.

        Parameters
        ----------
        frame_index: int
             Frame index to get

        Returns
        -------
        bytes
            The frame in bytes
        """
        raise NotImplementedError()

    def _get_tile_frames(self, frame_indices: Sequence[int]) -> Iterator[bytes]:
        return (self._get_tile_frame(frame_index) for frame_index in frame_indices)

    @cached_property
    def tiles(self) -> TileIndex:
        """Return tile index for image."""
        if self._datasets[0].tile_type == TileType.FULL:
            return FullTileIndex(self._datasets)
        else:
            return SparseTileIndex(self._datasets)

    @property
    def image_size(self) -> Size:
        """The pixel size of the image."""
        return self.tiles.image_size

    @property
    def tile_size(self) -> Size:
        """The pixel tile size of the image."""
        return self.tiles.tile_size

    @property
    def focal_planes(self) -> List[float]:
        """Focal planes available in the image defined in um."""
        return self.tiles.focal_planes

    @property
    def optical_paths(self) -> List[str]:
        """Optical paths available in the image."""
        return self.tiles.optical_paths

    @property
    def pixel_spacing(self) -> Optional[SizeMm]:
        """Size of the pixels in mm/pixel."""
        return self._datasets[0].pixel_spacing

    @property
    def photometric_interpretation(self) -> str:
        """Return photometric interpretation."""
        return self._datasets[0].photometric_interpretation

    @property
    def bits(self) -> int:
        """Return the number of bits stored for each sample."""
        return self._datasets[0].bits

    @property
    def samples_per_pixel(self) -> int:
        """Return samples per pixel (1 or 3)."""
        return self._datasets[0].samples_per_pixel

    @property
    def lossy_compressed(self) -> bool:
        return self._datasets[0].lossy_compressed

    @cached_property
    def image_coordinate_system(self) -> Optional[ImageCoordinateSystem]:
        """Return the image origin of the image data."""
        try:
            schema = ImageCoordinateSystemDicomSchema()
            return schema.load(self._datasets[0])
        except TypeError:
            return None

    @property
    def decoder(self) -> Decoder:
        return self._decoder

    def _get_encoded_tile(self, tile: Point, z: float, path: str) -> bytes:
        """
        Return bytes for tile.

        Parameters
        ----------
        tile: Point
            Tiles to get.
        z: float
            Z coordinate.
        path: str
            Optical path.

        Returns
        -------
        bytes
            Tile as bytes.
        """
        frame_index = self._get_frame_index(tile, z, path)
        if frame_index == -1:
            return self.blank_encoded_tile
        return self._get_tile_frame(frame_index)

    def _get_decoded_tile(self, tile: Point, z: float, path: str) -> Image:
        """
        Return Pillow image for tile.

        Parameters
        ----------
        tile: Point
            Tiles to get.
        z: float
            Z coordinate.
        path: str
            Optical path.

        Returns
        -------
        Image
            Tile as Pillow Image.
        """
        frame_index = self._get_frame_index(tile, z, path)
        if frame_index == -1:
            return self.blank_tile
        frame = self._get_tile_frame(frame_index)
        return self.decoder.decode(frame)

    def _get_frame_index(self, tile: Point, z: float, path: str) -> int:
        """
        Return frame index for tile.

        Raises WsiDicomOutOfBoundsError if tile, z, or path is not valid.

        Parameters
        ----------
        tile: Point
             Tile coordinate
        z: float
            Z coordinate
        path: str
            Optical identifier

        Returns
        -------
        int
            Tile frame index
        """
        tile_region = Region(position=tile, size=Size(0, 0))
        if not self.valid_tiles(tile_region, z, path):
            raise WsiDicomOutOfBoundsError(
                f"Tile region {tile_region}", f"plane {self.tiles.tiled_size}"
            )
        frame_index = self.tiles.get_frame_index(tile, z, path)
        return frame_index

    def is_sparse(self, tile: Point, z: float, path: str) -> bool:
        return self.tiles.get_frame_index(tile, z, path) == -1
